<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

/** @noinspection PhpFullyQualifiedNameUsageInspection */

require_once __DIR__ . '/app-translation.php';
require_once __DIR__ . '/classes/settings.php';
require_once __DIR__ . '/classes/csrf.php';
require_once __DIR__ . '/classes/manager.php';
require_once __DIR__ . '/classes/payments-manager.php';
require_once __DIR__ . '/classes/carriers-manager.php';
require_once __DIR__ . '/classes/customer-manager.php';
require_once __DIR__ . '/classes/entry-fields.php';
require_once __DIR__ . '/integration/shipping-integration.php';
require_once __DIR__ . '/integration/entry-fields-integration.php';
require_once __DIR__ . '/integration/integrations.php';

class Chex extends Module
{
    /** @var \Chex\CSRFToken */
    private $csrfToken;

    /** @var \Chex\Manager */
    private $manager;

    /** @var \Chex\Settings */
    private $settings;

    /** @var \Chex\Integrations */
    private $integrations;

    /**
     * Chex constructor.
     * @throws PrestaShopException
     */
    public function __construct()
    {
        $this->name = 'chex';
        $this->tab = 'administration';
        $this->version = '0.9.3';
        $this->author = 'DataKick';
        $this->need_instance = 0;
        $this->bootstrap = true;
        parent::__construct();
        $this->displayName = $this->l('Checkout Experience');
        $this->description = $this->l('One page checkout module');
        $this->ps_versions_compliancy = array('min' => '1.6', 'max' => '1.6.999');
        $this->controllers = ['checkout'];
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function install()
    {
        return (
            parent::install() &&
            $this->registerHooks() &&
            $this->installDb(true) &&
            $this->installTab()
        );
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function uninstall()
    {
        return (
            $this->removeTab() &&
            $this->uninstallDb(true) &&
            $this->unregisterHooks() &&
            parent::uninstall()
        );
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function registerHooks()
    {
        return $this->setupHooks([
            'displayHeader',
            'displayMobileHeader',
            'moduleRoutes',
            'displayAdminOrderTabOrder',
            'displayAdminOrderContentOrder',
        ]);
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function unregisterHooks()
    {
        return $this->setupHooks([]);
    }

    /**
     * @param array $hooks
     *
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function setupHooks($hooks)
    {
        $id = $this->id;
        $install = [];
        $delete = [];
        $aliases = Hook::getHookAliasList();
        foreach ($hooks as $hook) {
            $key = strtolower($hook);
            if (isset($aliases[$key])) {
                $hook = $aliases[$key];
                $key = strtolower($hook);
            }
            $install[$key] = $hook;
        }
        $sql = 'SELECT DISTINCT LOWER(h.name) AS `hook` FROM ' . _DB_PREFIX_ . 'hook h INNER JOIN ' . _DB_PREFIX_ . 'hook_module hm ON (h.id_hook = hm.id_hook) WHERE hm.id_module = ' . (int)$id;
        foreach (Db::getInstance()->executeS($sql) as $row) {
            $hook = $row['hook'];
            if (isset($install[$hook])) {
                unset($install[$hook]);
            } else {
                $delete[] = $hook;
            }
        }
        $ret = true;
        foreach ($install as $hook) {
            if (!$this->registerHook($hook)) {
                $ret = false;
            }
        }
        foreach ($delete as $hook) {
            $this->unregisterHook($hook);
        }
        return $ret;
    }

    /**
     * @param bool $create
     *
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function installDb($create) {
        if (! $create) {
            return true;
        }
        return $this->executeSqlScript('install');
    }

    /**
     * @param bool $drop
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function uninstallDb($drop) {
        if (! $drop) {
            return true;
        }
        return $this->executeSqlScript('uninstall', false);
    }

    /**
     * @param string $script
     * @param bool $check
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function executeSqlScript($script, $check=true) {
        $file = dirname(__FILE__) . '/sql/' . $script . '.sql';
        if (! file_exists($file)) {
            return false;
        }
        $sql = file_get_contents($file);
        if (! $sql) {
            return false;
        }
        $sql = str_replace(['PREFIX_', 'ENGINE_TYPE', 'CHARSET_TYPE'], [_DB_PREFIX_, _MYSQL_ENGINE_, 'utf8'], $sql);
        $sql = preg_split("/;\s*[\r\n]+/", $sql);
        foreach ($sql as $statement) {
            $stmt = trim($statement);
            if ($stmt) {
                try {
                    if (!Db::getInstance()->execute($stmt)) {
                        PrestaShopLogger::addLog("chex: sql script $script: $stmt: error");
                        if ($check) {
                            return false;
                        }
                    }
                } catch (Exception $e) {
                    PrestaShopLogger::addLog("chex: sql script $script: $stmt: exception");
                    if ($check) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    /**
     * @throws PrestaShopException
     */
    public function hookDisplayHeader()
    {
        $this->hookHeader();
    }

    /**
     * @throws PrestaShopException
     */
    public function hookDisplayMobileHeader()
    {
        $this->hookHeader();
    }

    /**
     * @throws PrestaShopException
     */
    public function hookMobileHeader()
    {
        $this->hookHeader();
    }

    /**
     * @param array $params
     *
     * @return string
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     * @throws SmartyException
     */
    public function hookDisplayAdminOrderTabOrder($params)
    {
        $values = $this->getOrderCustomFields($params);
        if ($values) {
            Context::getContext()->smarty->assign('chexCustomFields', $values);
            return $this->display(__FILE__, 'admin-order-tab.tpl');
        }
        return null;
    }

    /**
     * @param array $params
     * @return string|null
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     * @throws SmartyException
     */
    public function hookDisplayAdminOrderContentOrder($params)
    {
        $values = $this->getOrderCustomFields($params);
        if ($values) {
            Context::getContext()->smarty->assign('chexCustomFields', $values);
            return $this->display(__FILE__, 'admin-order-content.tpl');
        }
        return null;
    }

    /**
     * @param array $params
     *
     * @return array|null
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getOrderCustomFields($params)
    {
        static $sections = null;
        if (is_null($sections) && isset($params['order'])) {
            $sections = [];
            $cartId = (int)$params['order']->id_cart;
            $conn = Db::getInstance(_PS_USE_SQL_SLAVE_);
            $tables = [
                'chex_field_value_bool' => 'checkbox',
                'chex_field_value_int' => 'int',
                'chex_field_value_text' => 'text',
                'chex_field_value_longtext' => 'textarea',
            ];
            $sqls = [];
            foreach ($tables as $table => $type) {
                $sqls[] = (new DbQuery())
                    ->select('field, id_entity, value, section, "'.$type.'" as type')
                    ->from($table)
                    ->where('id_cart = ' . $cartId)
                    ->build();
            }
            $sql = implode("\nUNION ALL\n", $sqls);
            $rows = $conn->executeS($sql);
            if ($rows) {
                $fields = $this->getManager()->getEntryFieldsManager()->getIndexedFields();
                foreach ($rows as $row) {
                    $section = $row['section'];
                    $fieldId = $row['field'];
                    $entityId = (int)$row['id_entity'];
                    $sectionId = $section . '-' . $entityId;
                    $label = $fieldId;
                    $type = $row['type'];
                    if (isset($fields[$fieldId])) {
                        $field = $fields[$fieldId];
                        $label = $field['label'];
                        $type = $field['type'];
                    }

                    if (! isset($sections[$sectionId])) {
                        $sections[$sectionId] = [
                            'label' => $this->getSectionName($section, $entityId),
                            'fields' => [],
                        ];
                    }

                    $sections[$sectionId]['fields'][] = [
                        'field' => $label,
                        'fieldId' => $fieldId,
                        'type' => $type,
                        'value' => $row['value'],
                    ];


                }
            }
        }
        return is_null($sections) ? [] : $sections;
    }

    /**
     * @param string $type
     * @param int $id
     * @return string
     */
    private function getSectionName($type, $id)
    {
        switch ($type) {
            case 'confirm':
                return $this->l('Confirmation');
            case 'address':
                return sprintf($this->l('Address [id=%s]'), $id);
            case 'shipping':
                return $this->l('Shipping');
            case 'account':
                return $this->l('Customer account');
            default:
                return $type;
        }

    }

    /**
     * @throws PrestaShopException
     * @throws Exception
     */
    public function hookHeader()
    {
        $controller = $this->context->controller;
        if ($controller && $controller->php_self) {
            $link = $this->context->link;
            $page = $controller->php_self;
            if ($page === 'order' || $page === 'order-opc') {
                if ($this->useChex()) {
                    Tools::redirect($link->getModuleLink('chex', 'checkout', [], true));
                }
            }
        }
        $this->csrf();
    }

    /**
     * @return string|null
     * @throws PrestaShopException
     */
    public function getLicenseKey()
    {
        $license = $this->getSettings()->getLicense();
        if (!$license) {
            return null;
        }
        try {
            $decoded = base64_decode($license);
            $arr = json_decode($decoded, true);
            if (is_array($arr) && isset($arr['d']) && isset($arr['e']) && isset($arr['m']) && $arr['m'] === 'chex') {
                $expires = DateTime::createFromFormat('Y-m-d', $arr['e']);
                $now = new DateTime();
                if ($now < $expires) {
                    return $license;
                }
            }
        } catch (Exception $e) {
        }
        return null;
    }

    /**
     * returns true, if chex checkout should be used for current user
     *
     * @return boolean
     * @throws PrestaShopException
     */
    public function useChex()
    {
        $settings = $this->getSettings();
        if (!$this->getLicenseKey()) {
            return false;
        }
        if ($settings->isSandboxEnabled()) {
            $addresses = $settings->getSandboxAddresses();
            if (!in_array(Tools::getRemoteAddr(), $addresses)) {
                return false;
            }
        }
        return true;
    }

    /**
     * @return array
     */
    public function hookModuleRoutes()
    {
        return [
            'module-chex-checkout' => [
                'controller' => 'checkout',
                'rule' => 'checkout',
                'keywords' => [],
                'params' => [
                    'fc' => 'module',
                    'module' => $this->name,
                    'controller' => 'checkout',
                ]
            ],
        ];
    }

    /**
     * @throws PrestaShopException
     */
    public function getContent()
    {
        Tools::redirectAdmin($this->context->link->getAdminLink('AdminChexBackend'));
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function installTab()
    {
        $tab = new Tab();
        $tab->active = 1;
        $tab->class_name = 'AdminChexBackend';
        $tab->module = $this->name;
        $tab->id_parent = $this->getTabParent();
        $tab->name = array();
        foreach (Language::getLanguages(true) as $lang) {
            $tab->name[$lang['id_lang']] = 'Checkout Experience';
        }
        return $tab->add();
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function removeTab()
    {
        $tabId = Tab::getIdFromClassName('AdminChexBackend');
        if ($tabId) {
            $tab = new Tab($tabId);
            return $tab->delete();
        }
        return true;
    }

    /**
     * @return int
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getTabParent()
    {
        $parent = Tab::getIdFromClassName('AdminParentModules');
        if ($parent !== false) {
            return $parent;
        }
        return 0;
    }

    /**
     * @return array
     */
    public function getFrontTranslations()
    {
        $transl = new \Chex\AppTranslation($this);
        return $transl->getFrontTranslations();
    }

    /**
     * @return array
     */
    public function getBackTranslations()
    {
        $transl = new \Chex\AppTranslation($this);
        return $transl->getBackTranslations();
    }

    /**
     * @return \Chex\Settings
     */
    public function getSettings()
    {
        if (is_null($this->settings)) {
            $this->settings = new \Chex\Settings();
        }
        return $this->settings;
    }

    /**
     * @return \Chex\Manager
     */
    public function getManager()
    {
        if (is_null($this->manager)) {
            $this->manager = new \Chex\Manager($this);
        }
        return $this->manager;
    }

    /**
     * @return \Chex\CSRFToken
     * @throws Exception
     */
    public function csrf()
    {
        if (is_null($this->csrfToken)) {
            $this->csrfToken = new \Chex\CSRFToken($this->context->cookie, _COOKIE_KEY_);
        }
        return $this->csrfToken;
    }

    /**
     * @return \Chex\Integrations
     */
    public function getIntegrations()
    {
       if (is_null($this->integrations))  {
           $this->integrations = new \Chex\Integrations();
       }
       return $this->integrations;
    }

    /**
     * @return Context
     */
    public function getContext()
    {
        return $this->context;
    }

    /**
     * @param string $relative
     * @return string
     */
    public function getPath($relative)
    {
        $uri = rtrim($this->getPathUri(), '/');
        $rel = ltrim($relative, '/');
        return "$uri/$rel";
    }

}
