<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Carrier;
use Chex;
use Context;
use PrestaShopDatabaseException;
use PrestaShopException;

class CarriersManager
{
    /** @var Chex */
    private $module;

    /** @var array */
    private $carriers = null;

    /**
     * CarriersManager constructor.
     * @param Chex $module
     */
    public function __construct($module)
    {
        $this->module = $module;
    }

    /**
     * @return array|null
     * @throws PrestaShopException
     * @throws PrestaShopDatabaseException
     */
    public function getCarriers()
    {
        if (is_null($this->carriers)) {
            $this->carriers = [];
            $context = Context::getContext();
            $language = (int)$context->language->id;
            foreach (Carrier::getCarriers($language, false, false, false, null, Carrier::ALL_CARRIERS) as $carrier) {
                $id = (int)$carrier['id_carrier'];
                $logoUrl = null;
                if (file_exists(_PS_SHIP_IMG_DIR_ . $id . '.jpg')) {
                    $logoUrl = _THEME_SHIP_DIR_ . $id . '.jpg';
                }
                $def = [
                    'id' => $id,
                    'id_reference' => (int)$carrier['id_reference'],
                    'name' => $carrier['name'],
                    'delay' => $carrier['delay']
                ];
                if ($logoUrl) {
                    $def['logo'] = $logoUrl;
                }

                $this->carriers[$id] = $def;
            }
        }
        return $this->carriers;
    }

    /**
     * @param array $carrier
     *
     * @return array[]|null
     * @throws PrestaShopException
     */
    public function getCarrierEntryFields($carrier)
    {
        $module = $this->getShippingModuleForCarrier($carrier);
        if ($module) {
            return $module->getCarrierEntryFields($carrier);
        }
        return null;
    }

    /**
     * @param array $carrier
     * @return ShippingIntegration|null
     * @throws PrestaShopException
     */
    private function getShippingModuleForCarrier($carrier)
    {
        if (isset($carrier['external_module_name'])) {
            $module = $this->getShippingModule($carrier['external_module_name']);
            if ($module) {
                return $module;
            }
        }
        foreach ($this->getShippingModules() as $module) {
            if ($module->handlesCarrier($carrier)) {
                return $module;
            }
        }
        return null;
    }

    /**
     * @param string $name
     * @return ShippingIntegration|null
     * @throws PrestaShopException
     */
    private function getShippingModule($name)
    {
        $modules = $this->getShippingModules();
        if (!isset($modules[$name])) {
            return null;
        }
        return $modules[$name];
    }

    /**
     * @return ShippingIntegration[]
     * @throws PrestaShopException
     */
    private function getShippingModules()
    {
        return $this->module->getIntegrations()->getShippingIntegrations();
    }

    /**
     * @param int $carrierId
     * @param array $data
     * @return mixed
     * @throws PrestaShopException
     */
    public function processCustomData($carrierId, $data)
    {
        foreach ($this->getCarriers() as $carrier) {
            if ($carrier['id'] === $carrierId) {
                $module = $this->getShippingModuleForCarrier($carrier);
                $cart = Context::getContext()->cart;
                if (!$module) {
                    throw new PrestaShopException("Failed to find module that handles carrier $carrierId");
                }
                return $module->processData($cart, $carrier, $data);
            }
        }
        throw new PrestaShopException("Carrier with id $carrierId not found");
    }

    /**
     * @param int $carrierId
     * @param string $field
     * @param string $search
     * @return array
     * @throws PrestaShopException
     * @throws PrestaShopDatabaseException
     */
    public function getSelectOptions($carrierId, $field, $search)
    {
        foreach ($this->getCarriers() as $carrier) {
            if ($carrier['id'] === $carrierId) {
                $module = $this->getShippingModuleForCarrier($carrier);
                if (!$module) {
                    throw new PrestaShopException("Failed to find module that handles carrier $carrierId");
                }
                return $module->getSelectOptions($field, $search, $carrier);
            }
        }
        throw new PrestaShopException("Carrier with id $carrierId not found");
    }
}
