<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Cookie;
use Exception;
use PrestaShopException;

class CSRFToken
{
    /** @var string */
    private $token;

    /**
     * CSRFToken constructor.
     * @param Cookie $cookie
     * @param string $salt
     * @throws Exception
     */
    public function __construct($cookie, $salt)
    {
        $date = date('Ymd');
        $this->token = $this->getFromCookie($cookie, $salt, $date);
        if (is_null($this->token)) {
            $this->token = $this->createToken($salt, $date);
            $cookie->__set('chexToken', $this->token);
        }
    }

    /**
     * @return string
     */
    public function getToken()
    {
        return $this->token;
    }

    /**
     * @param string $token
     * @throws PrestaShopException
     */
    public function validate($token)
    {
        $passedToken = strtolower($token);
        $cookieToken = strtolower($this->token);
        if ($cookieToken != $passedToken) {
            throw new PrestaShopException("Invalid CSRF Token:\ncookie: $cookieToken\npassed: $passedToken\n");
        }
    }

    /**
     * @param string $salt
     * @param string $date
     * @return string
     */
    private function createToken($salt, $date)
    {
        $rand = self::getRandomData();
        return $rand . '-' . $this->getSignature($salt, $rand, $date);
    }

    /**
     * @param string $salt
     * @param string $data
     * @param string $date
     * @return string
     */
    private function getSignature($salt, $data, $date)
    {
        return md5($date . $salt . $data);
    }

    /**
     * @param Cookie $cookie
     * @param string $salt
     * @param string $date
     *
     * @return string|null
     *
     * @noinspection PhpUndefinedFieldInspection
     */
    private function getFromCookie(Cookie $cookie, $salt, $date)
    {
        if ($cookie->chexToken) {
            $token = $cookie->chexToken;
            $arr = explode('-', $token);
            if (count($arr) == 2) {
                $rand = $arr[0];
                $sign = $arr[1];
                $signature = $this->getSignature($salt, $rand, $date);
                if ($sign === $signature) {
                    return $token;
                }
            }
        }
        return null;
    }

    /**
     * @return string
     */
    public static function getRandomData()
    {
        try {
            if (function_exists('random_bytes')) {
                return bin2hex(random_bytes(16));
            }
            if (function_exists('openssl_random_pseudo_bytes')) {
                return bin2hex(openssl_random_pseudo_bytes(16));
            }
        } catch (Exception $e) {}
        return bin2hex(substr(hex2bin(sha1(uniqid('rd' . rand(), true))), 0, 16));
    }
}
