<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Context;
use Customer;
use Group;
use PrestaShopDatabaseException;
use PrestaShopException;
use Validate;
use Tools;
use Address;
use Configuration;
use Hook;
use Mail;

class CustomerManager
{
    /** @var Context */
    private $context;

    /**
     * CustomerManager constructor.
     * @param Context $context
     */
    public function __construct(Context $context)
    {
        $this->context = $context;
    }

    /**
     * @return array
     * @throws PrestaShopException
     */
    public function getCustomerData()
    {
        if (!$this->context->customer->isLogged(true)) {
            return $this->loggedOut();
        } else {
            return $this->loggedIn();
        }
    }

    /**
     * @return array
     * @throws PrestaShopException
     */
    public function loggedOut()
    {
        return [
            'type' => 'visitor',
            'showTax' => $this->showTax(),
        ];
    }

    /**
     * @return array
     * @throws PrestaShopException
     */
    public function loggedIn()
    {
        return $this->mapCustomer($this->context->customer);
    }

    /**
     * @param string $email
     * @return bool
     * @throws PrestaShopException
     */
    public function customerExists($email)
    {
        return Validate::isEmail($email) && Customer::customerExists($email, false, true);
    }

    /**
     * @param array $data
     *
     * @return array
     * @throws PrestaShopException
     * @throws PrestaShopDatabaseException
     *
     * @noinspection PhpUndefinedFieldInspection
     */
    public function createCustomer($data)
    {
        $customer = new Customer();

        $customer->active = 1;

        // get email
        if (!isset($data['email']) || !Validate::isEmail(self::toEmail($data['email']))) {
            throw new PrestaShopException("Invalid email");
        }
        $customer->email = self::toEmail($data['email']);

        // get firstname
        if (!isset($data['firstname'])) {
            throw new PrestaShopException("Invalid first name");
        }
        $customer->firstname = $data['firstname'];

        // get lastname
        if (!isset($data['lastname'])) {
            throw new PrestaShopException("Invalid last name");
        }
        $customer->lastname = $data['lastname'];

        // newsletter
        if (isset($data['newsletter'])) {
            $customer->newsletter = static::toBoolean($data['newsletter']);
            if ($customer->newsletter) {
                $customer->newsletter = true;
                $customer->ip_registration_newsletter = pSQL(Tools::getRemoteAddr());
                $customer->newsletter_date_add = pSQL(date('Y-m-d H:i:s'));
            }
        }

        // optin
        if (isset($data['optin'])) {
            $customer->optin = static::toBoolean($data['optin']);
        }

        // create either guest or customer account
        $isCustomer = isset($data['password']) && !empty($data['password']);
        $groups = [];
        if ($isCustomer) {
            // create customer
            if ($this->customerExists($customer->email)) {
                throw new PrestaShopException("Customer with this email already exists");
            }
            $customer->passwd = self::hashPassword($data['password']);
            $customer->is_guest = 0;
            $groups[] = (int)Configuration::get('PS_CUSTOMER_GROUP');
        } else {
            // create guest
            if (!Configuration::get('PS_GUEST_CHECKOUT_ENABLED')) {
                throw new PrestaShopException('Guest checkout is disabled');
            }
            $customer->passwd = self::hashPassword(sha1(_COOKIE_KEY_ . md5(time())));
            $customer->is_guest = 1;
            $groups[] = (int)Configuration::get('PS_GUEST_GROUP');
        }
        if (!$customer->add()) {
            throw new PrestaShopException("Failed to create customer");
        }
        $customer->cleanGroups();
        $customer->addGroups($groups);
        $customer->logged = 1;

        // fix environment
        $this->context->customer = $customer;
        $this->context->cookie->id_customer = (int)$customer->id;
        $this->context->cookie->customer_lastname = $customer->lastname;
        $this->context->cookie->customer_firstname = $customer->firstname;
        $this->context->cookie->passwd = $customer->passwd;
        $this->context->cookie->logged = 1;
        $this->context->cookie->email = $customer->email;
        $this->context->cookie->is_guest = !$isCustomer;
        $this->context->cart->secure_key = $customer->secure_key;

        $this->context->cart->update();
        $this->context->cart->autosetProductAddress();

        Hook::exec('actionCustomerAccountAdd', ['_POST' => $_POST, 'newCustomer' => $customer]);
        if ($isCustomer) {
            $this->sendConfirmationMail($customer);
        }

        return $this->mapCustomer($customer);
    }

    /**
     * @param array $address
     * @return array
     * @throws PrestaShopException
     */
    public function createAddress($address)
    {
        $customer = $this->context->customer;
        if (!$customer->isLogged(true)) {
            throw new PrestaShopException("Can't create address without customer");
        }
        $addr = new Address();
        $addr->id_customer = (int)$customer->id;
        $addr->id_country = (int)$address['country'];
        $addr->id_state = (int)$address['state'];
        $addr->company = $address['company'];
        $addr->lastname = $address['lastname'];
        $addr->firstname = $address['firstname'];
        $addr->address1 = $address['address1'];
        $addr->address2 = $address['address2'];
        $addr->postcode = $address['postcode'];
        $addr->city = $address['city'];
        $addr->phone = $address['phone'];
        $addr->phone_mobile = $address['mobilePhone'];
        $addr->vat_number = $address['vatNumber'];
        $addr->dni = $address['dni'];
        $addr->alias = self::generateAlias($address);
        if (!$addr->save() || !Validate::isLoadedObject($addr)) {
            throw new PrestaShopException("Can't save address");
        }
        return self::mapAddress($addr);
    }

    /**
     * @return array
     * @throws PrestaShopException
     */
    public function getAddresses()
    {
        $customer = $this->context->customer;
        if ($customer->isLogged(true)) {
            return array_map(['\Chex\CustomerManager', 'mapAddress'], $customer->getAddresses($this->context->language->id));
        } else {
            return [];
        }
    }

    /**
     * @param array $addr
     *
     * @return string
     */
    public static function generateAlias($addr)
    {
        $alias = mb_substr(implode(', ', [
            $addr['address1'],
            $addr['city'],
            $addr['postcode'],
        ]), 0, 32);

        return preg_replace(Tools::cleanNonUnicodeSupport('/[<>={}]+/u'), ' ', $alias);
    }

    /**
     * @return array
     */
    public static function getMapping()
    {
        return [
            'id_address' => 'id',
            'id_country' => 'country',
            'id_state' => 'state',
            'company' => 'company',
            'lastname' => 'lastname',
            'firstname' => 'firstname',
            'address1' => 'address1',
            'address2' => 'address2',
            'postcode' => 'postcode',
            'city' => 'city',
            'phone' => 'phone',
            'phone_mobile' => 'mobilePhone',
            'vat_number' => 'vatNumber',
        ];
    }

    /**
     * @param Customer $customer
     *
     * @return array
     * @throws PrestaShopException
     */
    private function mapCustomer(Customer $customer)
    {
        return [
            'type' => $customer->isGuest() ? 'guest' : 'customer',
            'id' => (int)$customer->id,
            'email' => $customer->email,
            'firstname' => $customer->firstname,
            'lastname' => $customer->lastname,
            'addresses' => $this->getAddresses(),
            'showTax' => $this->showTax(),
        ];
    }

    /**
     * @return bool
     * @throws PrestaShopException
     */
    private function showTax()
    {
        $group = Group::getCurrent();
        return Group::getPriceDisplayMethod($group->id) == 0;
    }

    /**
     * @param array|Address $addr
     * @return array
     */
    private static function mapAddress($addr)
    {
        $ret = [];
        if (!is_array($addr)) {
            $addr = (array)$addr;
            $addr['id_address'] = $addr['id'];
        }
        foreach (self::getMapping() as $key => $val) {
            $value = strpos($key, 'id_') === 0 ? (int)$addr[$key] : $addr[$key];
            $ret[$val] = $value;
        }
        return $ret;
    }

    /**
     * @param Customer $customer
     *
     * @return void
     * @throws PrestaShopException
     */
    private function sendConfirmationMail(Customer $customer): void
    {
        if (!Configuration::get('PS_CUSTOMER_CREATION_EMAIL')) {
            return;
        }

        Mail::Send(
            $this->context->language->id,
            'account',
            Mail::l('Welcome!'),
            [
                '{firstname}' => $customer->firstname,
                '{lastname}' => $customer->lastname,
                '{email}' => $customer->email,
                '{passwd}' => '*******',
            ],
            $customer->email,
            $customer->firstname . ' ' . $customer->lastname
        );
    }

    /**
     * @param string $str
     * @return string
     */
    private static function hashPassword($str)
    {
        if (is_callable(['Tools', 'hash'])) {
            return Tools::hash($str);
        } else {
            return Tools::encrypt($str);
        }
    }

    /**
     * @param string $str
     * @return string
     */
    private static function toEmail($str)
    {
        if (is_callable(['Tools', 'convertEmailToIdn'])) {
            return Tools::convertEmailToIdn($str);
        }
        return $str;
    }

    /**
     * @param string $email
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function resetPasswordRequest($email)
    {
        $customer = new Customer();
        $customer->getByEmail($email);
        if (! Validate::isLoadedObject($customer)) {
            throw new PrestaShopException('Customer account not found');
        }

        if (! $customer->active) {
            throw new PrestaShopException('You cannot regenerate the password for this account.');
        }

        $minutes = (int) Configuration::get('PS_PASSWD_TIME_FRONT');
        if ((strtotime($customer->last_passwd_gen.'+'. $minutes . ' minutes') - time()) > 0) {
            throw new PrestaShopException("You have to wait $minutes between password reset attempts");
        }

        $mailParams = [
            '{email}'     => $customer->email,
            '{lastname}'  => $customer->lastname,
            '{firstname}' => $customer->firstname,
            '{url}'       => $this->context->link->getPageLink('password', true, null, 'token='.$customer->secure_key.'&id_customer='.(int) $customer->id),
        ];
        if (! Mail::Send($this->context->language->id, 'password_query', Mail::l('Password query confirmation'), $mailParams, $customer->email, $customer->firstname.' '.$customer->lastname)) {
            throw new PrestaShopException('An error occurred while sending the email.');
        }
        return true;
    }

    /**
     * @param mixed $val
     * @return bool
     */
    private static function toBoolean($val)
    {
        if (is_null($val)) {
            return false;
        }
        if (is_bool($val)) {
            return (bool)$val;
        }
        if (is_string($val)) {
            $val = strtolower($val);
            if ($val === 'true' || $val === 'yes') {
                return true;
            }
            if ($val === 'false' || $val === 'no') {
                return false;
            }
        }
        if (is_numeric($val)) {
            return (bool)(int)$val;
        }
        return (bool)$val;
    }

}
