<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Chex;
use Configuration;
use CMS;
use Db;
use DbQuery;
use PrestaShopDatabaseException;
use PrestaShopException;
use Translate;
use Validate;

class EntryFields
{
    const TYPE_PASSWORD = 'password';
    const TYPE_CHECKBOX = 'checkbox';
    const TYPE_TEXT = 'text';
    const TYPE_TEXTAREA = 'textarea';
    const TYPE_SELECT = 'select';
    const TYPE_COUNTRY = 'country';
    const TYPE_STATE = 'state';
    const TYPE_TEL = 'tel';
    const TYPE_NUMBER = 'number';
    const TYPE_EMAIL = 'email';

    /** @var Chex */
    private $module;

    /** @var array  */
    private $fields = null;

    /** @var array */
    private $fieldPreferences;

    /**
     * EntryFields constructor.
     *
     * @param Chex $module
     */
    public function __construct(Chex $module)
    {
        $this->module = $module;
    }

    /**
     * @param string $str
     * @return string
     */
    public function l($str)
    {
        return $this->module->l($str, 'entry-fields');
    }

    /**
     * @param bool $includeHidden
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getAccountLayout($includeHidden = false)
    {
        $layout = [
            ['account.password'],
            ['address.company', 'address.vatNumber'],
            ['address.dni'],
            ['address.firstname', 'address.lastname'],
            ['address.address1'],
            ['address.address2'],
            ['address.city', 'address.postcode'],
            ['address.country', 'address.state'],
            ['address.phone', 'address.mobilePhone'],
            ['account.optin'],
            ['account.newsletter'],
        ];

        foreach ($this->getCustomFields('account') as $field) {
            $layout[] = [ $field['id'] ];
        }
        foreach ($this->getCustomFields('address') as $field) {
            $layout[] = [ $field['id'] ];
        }
        return $includeHidden ? $layout : $this->filterHiddenFields($layout);
    }

    /**
     * @param bool $includeHidden
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getAddressLayout($includeHidden = false)
    {
        $layout = [
            ['address.firstname', 'address.lastname'],
            ['address.company', 'address.vatNumber'],
            ['address.dni'],
            ['address.address1'],
            ['address.address2'],
            ['address.city', 'address.postcode'],
            ['address.country', 'address.state'],
            ['address.phone', 'address.mobilePhone']
        ];
        foreach ($this->getCustomFields('address') as $field) {
            $layout[] = [ $field['id'] ];
        }
        return $includeHidden ? $layout : $this->filterHiddenFields($layout);
    }

    /**
     * @param bool $includeHidden
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getConfirmationLayout($includeHidden = false)
    {
        $layout = [
            ['confirm.message'],
            ['confirm.tos']
        ];
        foreach ($this->getCustomFields('confirm') as $field) {
            $layout[] = [ $field['id'] ];
        }

        return $includeHidden ? $layout : $this->filterHiddenFields($layout);
    }

    /**
     * @param bool $includeHidden
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getFields($includeHidden = false)
    {
        if ($includeHidden) {
            return array_values($this->getIndexedFields());
        }

        $ret = [];
        foreach ($this->getIndexedFields() as $id => $field) {
            if (! $this->isHiddenField($id)) {
                $ret[] = $field;
            }
        }
        return $ret;
    }

    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getIndexedFields()
    {
        if (is_null($this->fields)) {
            $this->fields = $this->initFields();
        }
        return $this->fields;
    }

    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function initFields()
    {
        $phoneRequired = (bool)Configuration::get('PS_ONE_PHONE_AT_LEAST');
        $newsletter = Configuration::get('PS_CUSTOMER_NWSL');
        $optin = (bool)Configuration::get('PS_CUSTOMER_OPTIN');
        $tosPage = $this->getTosPage();
        $fields = [
            [
                'id' => 'account.password',
                'label' => $this->l('Password'),
                'placeholder' => $this->l('Please choose your password'),
                'required' => !Configuration::get('PS_GUEST_CHECKOUT_ENABLED'),
                'type' => static::TYPE_PASSWORD,
                'validate' => 'password'
            ],
            [
                'id' => 'account.optin',
                'label' => $this->l('Receive offers from our partners'),
                'required' => false,
                'hidden' => !$optin,
                'conditional' => $this->l('displayed only when you enable optin in Customer Preferences'),
                'type' => static::TYPE_CHECKBOX,
                'defaultValue' => false,
            ],
            [
                'id' => 'account.newsletter',
                'label' => $this->l('Sign up for our newsletter'),
                'required' => false,
                'hidden' => !$newsletter,
                'conditional' => $this->l('displayed only when you enable newsletter in Customer Preferences'),
                'type' => static::TYPE_CHECKBOX,
                'defaultValue' => false,
            ],
            [
                'id' => 'address.company',
                'label' => $this->l('Company name'),
                'placeholder' => $this->l('Please enter company name'),
                'required' => false,
                'type' => static::TYPE_TEXT,
                'validate' => 'genericName',
            ],
            [
                'id' => 'address.vatNumber',
                'label' => $this->l('VAT number'),
                'placeholder' => $this->l('Please enter your vat number'),
                'required' => false,
                'type' => static::TYPE_TEXT,
                'validate' => 'vatNumber',
            ],
            [
                'id' => 'address.firstname',
                'name' => 'first_name',
                'label' => $this->l('First name'),
                'placeholder' => $this->l('Please enter your first name'),
                'required' => true,
                'type' => static::TYPE_TEXT,
                'validate' => 'name',
            ],
            [
                'id' => 'address.lastname',
                'label' => $this->l('Last name'),
                'placeholder' => $this->l('Please enter your last name'),
                'required' => true,
                'type' => static::TYPE_TEXT,
                'validate' => 'name',
            ],
            [
                'id' => 'address.address1',
                'label' => $this->l('Address line 1'),
                'placeholder' => $this->l('Enter your address'),
                'required' => true,
                'type' => static::TYPE_TEXT,
                'validate' => 'address',
            ],
            [
                'id' => 'address.address2',
                'label' => $this->l('Address line 2'),
                'placeholder' => $this->l('Enter your address'),
                'required' => false,
                'type' => static::TYPE_TEXT,
                'validate' => 'address',
            ],
            [
                'id' => 'address.postcode',
                'label' => $this->l('Postal code'),
                'placeholder' => $this->l('Please enter postal or zip code'),
                'required' => true,
                'type' => static::TYPE_TEXT,
                'validate' => 'postcode',
            ],
            [
                'id' => 'address.country',
                'label' => $this->l('Country'),
                'placeholder' => $this->l('Please select country'),
                'required' => true,
                'type' => static::TYPE_COUNTRY,
            ],
            [
                'id' => 'address.state',
                'label' => $this->l('State'),
                'placeholder' => $this->l('Please select state'),
                'required' => true,
                'type' => static::TYPE_STATE,
                'conditional' => $this->l('displayed only for countries with states'),
            ],
            [
                'id' => 'address.city',
                'label' => $this->l('City'),
                'placeholder' => $this->l('City'),
                'required' => true,
                'type' => static::TYPE_TEXT,
                'validate' => 'city',
            ],
            [
                'id' => 'address.phone',
                'label' => $this->l('Home phone'),
                'placeholder' => $this->l('Home phone number'),
                'required' => $phoneRequired,
                'requiredMessage' => $phoneRequired ? $this->l('Either home or mobile phone number must be entered') : '',
                'type' => static::TYPE_TEXT,
                'icon' => 'phone',
                'validate' => 'phoneNumber',
            ],
            [
                'id' => 'address.mobilePhone',
                'label' => $this->l('Mobile phone'),
                'placeholder' => $this->l('Mobile phone number'),
                'required' => $phoneRequired,
                'requiredMessage' => $phoneRequired ? $this->l('Either home or mobile phone number must be entered') : '',
                'type' => static::TYPE_TEXT,
                'icon' => 'mobile',
                'validate' => 'phoneNumber',
            ],
            [
                'id' => 'address.dni',
                'label' => $this->l('Identification number'),
                'placeholder' => $this->l('DNI / NIF / NI'),
                'required' => false,
                'type' => static::TYPE_TEXT,
                'validate' => 'dni',
            ],
            [
                'id' => 'confirm.message',
                'label' => $this->l('Leave a message'),
                'placeholder' => $this->l('If you would like to add a comment about your order, please write it here'),
                'required' => false,
                'type' => static::TYPE_TEXTAREA,
            ],
            [
                'id' => 'confirm.tos',
                'label' => Translate::ppTags(
                    $this->l('I agree to the [1]Terms of Service[/1] and will adhere to them unconditionally'),
                    ['<a href="' . $tosPage . '" target="tos">']
                ),
                'required' => true,
                'hidden' => !$tosPage,
                'conditional' => $this->l('displayed only when you choose CMS page in Order Preferences'),
                'type' => static::TYPE_CHECKBOX,
            ],
        ];

        // carrier entry fields
        $carrierManager = $this->module->getManager()->getCarrierManager();
        foreach ($carrierManager->getCarriers() as $carrier) {
            $entryFields = $carrierManager->getCarrierEntryFields($carrier);
            if ($entryFields) {
                foreach ($entryFields as $entryField) {
                    $field = $entryField['id'];
                    $entryField['id'] = 'carrier' . (int)$carrier['id'] . '.' . $field;
                    if (isset($entryField['select']) && $entryField['select']['type'] === 'ajax') {
                        $entryField['select']['ajax'] = [
                            'type' => 'carrier',
                            'carrierId' => (int)$carrier['id'],
                            'field' => $field,
                        ];
                    }
                    $fields[] = $entryField;
                }
            }
        }

        // integration entry fields
        foreach ($this->module->getIntegrations()->getEntryFieldsIntegrations() as $integration) {
            foreach ($integration->getEntryFields() as $entryField) {
                $entryField['custom'] = $integration->getId();
                if (isset($entryField['select']) && $entryField['select']['type'] === 'ajax') {
                    $field = $entryField['id'];
                    $entryField['select']['ajax'] = [
                        'type' => 'entryField',
                        'integration' => $integration->getId(),
                        'field' => $field,
                    ];
                }
                $fields[] = $entryField;
            }
        }

        // custom entry fields
        $sql = (new DbQuery())
            ->select('cf.*')
            ->from('chex_custom_field', 'cf');
        foreach (Db::getInstance()->executeS($sql) as $row) {
            $field = [
                'id' => $row['id'],
                'label' => $row['label'],
                'placeholder' => $row['placeholder'],
                'required' => !!(int)$row['required'],
                'type' => $row['type'],
                'custom' => 'chex',
            ];
            if (isset($row['validate']) && $row['validate'] !== '') {
                $field['validate'] = $row['validate'];
            }

            if (isset($row['default_value']) && $row['default_value'] !== '') {
                $field['defaultValue'] = $row['default_value'];
            }

            if (isset($row['required_message']) && $row['required_message'] !== '') {
                $field['requiredMessage'] = $row['required_message'];
            }

            $fields[] = $field;
        }

        $preferences = $this->getFieldsPreferences();
        $ret = [];
        foreach ($fields as $field) {
            $id = $field['id'];
            $module = $field['custom'] ?? 'chex';
            $fullId = $module . '.' . $id;
            if (isset($preferences[$fullId])) {
                $pref = $preferences[$fullId];
                if (!(isset($field['hidden']) && $field['hidden']) && !$pref['active']) {
                    $field['hidden'] = true;
                }
            }
            $ret[$id] = $field;
        }
        return $ret;
    }

    /**
     * @param string $id
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getField($id)
    {
        $index = $this->getIndexedFields();
        $ret = $index[$id];
        if (! $ret) {
            throw new PrestaShopException("Field $id not found");
        }
        return $ret;
    }

    /**
     * @return string
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getTosPage()
    {
        $cmsId = (int)Configuration::get('PS_CONDITIONS_CMS_ID');
        $context = $this->module->getContext();
        $cms = new CMS($cmsId, $context->language->id);
        if (Validate::isLoadedObject($cms)) {
            $link = $context->link->getCMSLink($cms, $cms->link_rewrite, (bool)Configuration::get('PS_SSL_ENABLED'));
            if (!strpos($link, '?')) {
                $link .= '?content_only=1';
            } else {
                $link .= '&content_only=1';
            }
            return $link;
        }
        return '';
    }

    /**
     * @param CarriersManager $carrierManager
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getCarrierLayouts(CarriersManager $carrierManager)
    {
        $layouts = [];
        foreach ($carrierManager->getCarriers() as $carrier) {
            $entryFields = $carrierManager->getCarrierEntryFields($carrier);
            $carrierId = 'carrier' . (int)$carrier['id'];
            if ($entryFields) {
                $layout = [];
                foreach ($entryFields as $field) {
                    $fieldId = $carrierId . '.' . $field['id'];
                    $layout[] = [$fieldId];
                }
                $layouts[$carrierId] = $layout;
            }
        }
        return $layouts;
    }

    /**
     * @param string $section
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getCustomFields($section)
    {
        $ret = [];
        foreach ($this->getIndexedFields() as $field) {
            if (isset($field['custom'])) {
                $id = explode('.', $field['id']);
                if ($id[0] === $section) {
                    $ret[] = $field;
                }
            }
        }
        return $ret;
    }

    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getFieldsPreferences()
    {
        if (is_null($this->fieldPreferences)) {
            $this->fieldPreferences = [];
            $rows = Db::getInstance()->executeS(
                (new DbQuery())
                    ->select('*')
                    ->from('chex_field_preference')
            );
            foreach ($rows as $entry) {
                $id = $entry['module'] . '.' . $entry['field'];
                $pref = array_merge($this->getDefaultPreferences(), [
                    'active' => !!$entry['active']
                ]);
                $this->fieldPreferences[$id] = $pref;
            }
        }
        return $this->fieldPreferences;
    }

    /**
     * @param string $id
     * @return array|mixed
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getFieldPreferences($id)
    {
        $preferences = $this->getFieldsPreferences();
        if (isset($preferences[$id])) {
            return $preferences[$id];
        }
        return $this->getDefaultPreferences();
    }

    /**
     * @return array
     */
    public function getDefaultPreferences()
    {
        return [
            'active' => true
        ];
    }

    /**
     * @param array $layout
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function filterHiddenFields($layout)
    {
        return array_reduce($layout, function($layout, $row) {
            $newRow = array_reduce($row, function($row, $id) {
                if (! $this->isHiddenField($id)) {
                    $row[] = $id;
                }
                return $row;
            }, []);
            if ($newRow) {
                $layout[] = $newRow;
            }
            return $layout;
        }, []);
    }

    /**
     * @param string $id
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function isHiddenField($id)
    {
        $field = $this->getField($id);
        return isset($field['hidden']) && $field['hidden'];
    }

    /**
     * @param string $field
     * @param boolean $active
     * @throws PrestaShopException
     */
    public function enableField($field, $active)
    {
        $parts = explode('.', $field);
        if (count($parts) >= 2) {
            $this->fieldPreferences = null;
            $module = pSQL(array_shift($parts));
            $id = pSQL(implode('.', $parts));
            $active = (int)$active;
            $sql = (
                "INSERT INTO "._DB_PREFIX_."chex_field_preference(`module`, `field`, `active`)\n" .
                "VALUES ('$module', '$id', $active)\n" .
                "ON DUPLICATE KEY UPDATE `module` = VALUES(`module`), `active` = VALUES(`active`)"
            );
            Db::getInstance()->execute($sql);
        }
    }

    /**
     * @param array $field
     * @throws PrestaShopException
     */
    public function saveField($field)
    {
        $required = ['id', 'required', 'type', 'label'];
        foreach ($required as $key) {
            if (! array_key_exists($key, $field)) {
                throw new PrestaShopException("Required property $key not provided");
            }
        }
        $id = $field['id'];
        $required = $field['required'];
        $type = $field['type'];
        $label = $field['label'];
        $custom = array_key_exists('custom', $field) ? $field['custom'] : null;
        $defaultValue = array_key_exists('defaultValue', $field) ? $field['defaultValue'] : null;
        $validate = array_key_exists('validate', $field) ? $field['validate'] : null;
        $requiredMessage = array_key_exists('requiredMessage', $field) ? $field['requiredMessage'] : null;
        $placeholder = array_key_exists('placeholder', $field) ? $field['placeholder'] : null;
        $icon = array_key_exists('icon', $field) ? $field['icon'] : null;
        // $conditional = array_key_exists('conditional', $field) ? $field['conditional'] : null;
        // $select = array_key_exists('select', $field) ? $field['select'] : null;

        $this->fieldPreferences = null;
        $this->fields = null;

        if ($custom === 'chex') {
            Db::getInstance()->insert('chex_custom_field', [
                'id' => pSQL($id),
                'required' => (int)(bool)$required,
                'type' => pSQL($type),
                'label' => pSQL($label, true),
                'placeholder' => pSQL($placeholder),
                'required_message' => pSQL($requiredMessage),
                'default_value' => pSQL($defaultValue),
                'validate' => pSQL($validate),
                'icon' => pSQL($icon),
            ], false, false, Db::ON_DUPLICATE_KEY);
        } else {
            $this->extendField();
        }
    }

    /**
     * @throws PrestaShopException
     */
    private function extendField()
    {
        throw new PrestaShopException('At the moment only custom fields can be edited');
    }

    /**
     * @param string $id
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function deleteField($id)
    {
        $id = pSQL($id);
        $cond = "`id` = '$id'";
        $exists = (new DbQuery())
            ->select('1')
            ->from('chex_custom_field')
            ->where($cond);
        if (! Db::getInstance()->getRow($exists)) {
            throw new PrestaShopException('Custom field not found');
        }
        return Db::getInstance()->delete('chex_custom_field', $cond);
    }


    /**
     * @param array $field
     * @return mixed
     * @throws PrestaShopException
     */
    public function getFieldDefaultValue($field)
    {
        if (isset($field['defaultValue'])) {
            return $field['defaultValue'];
        }
        switch ($field['type']) {
            case static::TYPE_PASSWORD:
            case static::TYPE_TEXT:
            case static::TYPE_TEXTAREA:
            case static::TYPE_SELECT:
            case static::TYPE_EMAIL:
            case static::TYPE_TEL:
                return null;
            case static::TYPE_CHECKBOX:
                return false;
            case static::TYPE_COUNTRY:
            case static::TYPE_STATE:
            case static::TYPE_NUMBER:
                return 0;
        }
        throw new PrestaShopException("Invalid field type: " . $field['type']);
    }

}
