<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Cart;
use CartRule;
use Chex;
use Configuration;
use Country;
use Currency;
use Exception;
use PrestaShopDatabaseException;
use PrestaShopException;
use State;
use Context;
use Validate;
use Zone;
use Tools;
use Db;
use Address;
use Cache;
use Module;
use VatNumber;
use Hook;

class Manager
{
    /** @var Chex */
    public $module;

    /** @var Context */
    private $context;

    /** @var PaymentsManager */
    private $paymentsManager;

    /** @var CarriersManager */
    private $carriersManager;

    /** @var CustomerManager  */
    private $customerManager;

    /** @var EntryFields  */
    private $fieldsManager;

    /**
     * Manager constructor.
     * @param Chex $module
     */
    public function __construct(Chex $module)
    {
        $this->module = $module;
        $this->context = Context::getContext();
        $this->paymentsManager = new PaymentsManager();
        $this->carriersManager = new CarriersManager($module);
        $this->customerManager = new CustomerManager($this->context);
        $this->fieldsManager = new EntryFields($module);
    }

    /**
     * @return CustomerManager
     */
    public function getCustomerManager()
    {
        return $this->customerManager;
    }

    /**
     * @return CarriersManager
     */
    public function getCarrierManager()
    {
        return $this->carriersManager;
    }

    /**
     * @param bool $includeHidden
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getEntryFields($includeHidden = false)
    {
        return $this->fieldsManager->getFields($includeHidden);
    }

    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     * @throws Exception
     */
    public function getInitialData()
    {
        $countries = $this->getCountries();
        $settings = $this->module->getSettings();
        $link = $this->context->link;
        return [
            'version' => $this->module->version,
            'license' => $this->module->getLicenseKey(),
            'csrf' => $this->module->csrf()->getToken(),
            'translations' => $this->module->getFrontTranslations(),
            'urls' => [
                'api' => $link->getModuleLink('chex', 'checkout', [], true),
                'auth' => $link->getPageLink('authentication', true)
            ],
            'carriers' => $this->carriersManager->getCarriers(),
            'countries' => $countries['countries'],
            'states' => $countries['states'],
            'zones' => $countries['zones'],
            'flow' => $this->getFlow(),
            'fields' => $this->fieldsManager->getFields(),
            'cartPosition' => $settings->getCartPosition(),
            'addressLayout' => $this->fieldsManager->getAddressLayout(),
            'accountLayout' => $this->fieldsManager->getAccountLayout(),
            'confirmLayout' => $this->fieldsManager->getConfirmationLayout(),
            'carrierLayouts' => $this->fieldsManager->getCarrierLayouts($this->getCarrierManager()),
            'displayTaxInCart' => !!Configuration::get('PS_TAX_DISPLAY'),
            'guestCheckoutEnabled' => !!Configuration::get('PS_GUEST_CHECKOUT_ENABLED'),
            'showCarrierLogo' => $settings->showCarrierLogo(),
            'showPaymentLogo' => $settings->showPaymentLogo(),
            'allowFloatingCart' => $settings->allowFloatingCart(),
            'data' => $this->getData()
        ];
    }

    /**
     * @param string $vatNumber
     * @return string | null
     * @throws PrestaShopException
     */
    public function checkVatNumber($vatNumber)
    {
        if ($this->allowTaxExempt() && Configuration::get('VATNUMBER_CHECKING')) {
            try {
                $errors = VatNumber::WebServiceCheck($vatNumber);
                if (count($errors)) {
                    return $errors[0];
                }
            } catch (Exception $e) {
                return 'Unknown error when checking VAT number';
            }
        }
        return null;
    }

    /**
     * @return string | null
     * @throws PrestaShopException
     */
    public function checkAvailability()
    {
        $cart = Context::getContext()->cart;
        $product = $cart->checkQuantities(true);
        if (is_array($product)) {
            $productName = $product['name'];
            if (isset($product['attributes_small']) && $product['attributes_small']) {
                $productName .= " (" . $product['attributes_small'] . ")";
            }
            $stockQuantity = (int)$product['stock_quantity'];
            if ($stockQuantity > 0) {
                return sprintf($this->l('Product %s is no longer available in requested quantity. We have only %s items left'), $productName, $stockQuantity);
            } else {
                return sprintf($this->l('Product %s is no longer available for sale'), $productName);
            }
        }

        return null;
    }

    /**
     * @return bool
     * @throws PrestaShopException
     */
    private function allowTaxExempt()
    {
        if (Module::isEnabled('vatnumber')) {
            /** @noinspection PhpIncludeInspection */
            require_once _PS_MODULE_DIR_ . '/vatnumber/vatnumber.php';
            return true;
        }
        return false;
    }

    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getCountries()
    {
        $lang = $this->context->language->id;
        $states = [];
        $countries = [];
        $zones = [];

        $taxExempt = $this->allowTaxExempt();
        $vatCountry = $taxExempt ? Configuration::get('VATNUMBER_COUNTRY') : 0;

        $allCountries = $this->module->getSettings()->getCountrySelectorMode() === 'all';

        foreach (Country::getCountries($lang) as $country) {
            $id = (int)$country['id_country'];
            $active = !!$country['active'];
            if ($active || $allCountries) {
                $data = [
                    'id' => $id,
                    'name' => $country['country'],
                    'zone' => (int)$country['id_zone'],
                    'active' => $active,
                    'iso' => $country['iso_code'],
                ];
                if ($taxExempt && $id != $vatCountry && VatNumber::isApplicable($id)) {
                    $data['vatExempt'] = true;
                }
                if (isset($country['states'])) {
                    $data['states'] = [];
                    foreach ($country['states'] as $state) {
                        $id = (int)$state['id_state'];
                        $data['states'][] = $id;
                        $states[$id] = [
                            'name' => $state['name'],
                            'zone' => (int)$state['id_zone'],
                            'active' => !!$state['active']
                        ];
                    }
                }
                if (!(int)$country['need_zip_code']) {
                    $data['noPostCode'] = true;
                }
                if ((int)$country['need_identification_number']) {
                    $data['dniRequired'] = true;
                }
                if ($country['zip_code_format']) {
                    $data['postCodeFormat'] = $country['zip_code_format'];
                }
                if ((int)$country['display_tax_label']) {
                    $data['taxLabel'] = true;
                }
                $countries[] = $data;
            }
        }

        foreach (Zone::getZones() as $zone) {
            $id = (int)$zone['id_zone'];
            $zones[$id] = !!$zone['active'];
        }

        return [
            'countries' => $countries,
            'states' => $states,
            'zones' => $zones,
        ];
    }

    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getCart()
    {
        $settings = $this->module->getSettings();
        $link = $this->context->link;
        $cart = $this->context->cart;
        Cache::clean('getContextualValue_*');
        $summary = $cart->getSummaryDetails();

        $virtual = true;
        $products = [];
        $imageType = $settings->getImageType();
        foreach ($summary['products'] as $prod) {
            if (!$prod['is_virtual']) {
                $virtual = false;
            }
            $products[] = [
                'id' => (int)$prod['id_product'],
                'combinationId' => (int)$prod['id_product_attribute'],
                'name' => $prod['name'],
                'attributes' => $prod['attributes_small'] ?? '',
                'image' => $link->getImageLink($prod['link_rewrite'], $prod['id_image'], $imageType),
                'quantity' => (int)$prod['cart_quantity'],
                'price' => self::price($prod['total'], $prod['total_wt']),
                'isGift' => false
            ];
        }
        foreach ($summary['gift_products'] as $prod) {
            $products[] = [
                'id' => (int)$prod['id_product'],
                'combinationId' => (int)$prod['id_product_attribute'],
                'name' => $prod['name'],
                'attributes' => $prod['attributes_small'],
                'image' => $link->getImageLink($prod['link_rewrite'], $prod['id_image'], $imageType),
                'quantity' => (int)$prod['cart_quantity'],
                'price' => self::price($prod['total'], $prod['total_wt']),
                'isGift' => true
            ];
        }
        $discounts = [];
        foreach ($summary['discounts'] as $disc) {
            $discounts[] = [
                'id' => (int)$disc['id_cart_rule'],
                'name' => $disc['name'],
                'price' => self::price($disc['value_tax_exc'], $disc['value_real']),
            ];
        }
        if (Configuration::get('CHEX_ALL_PRODUCTS_ARE_VIRTUAL')) {
            $virtual = true;
        }

        $currency = Currency::getCurrencyInstance((int) $cart->id_currency);
        $minimalPurchase = Tools::convertPrice((float) Configuration::get('PS_PURCHASE_MINIMUM'), $currency);

        $ret = [
            'virtual' => $virtual,
            'products' => $products,
            'discounts' => $discounts,
            'productsPrice' => self::price($summary['total_products'], $summary['total_products_wt']),
            'shippingPrice' => self::price($summary['total_shipping_tax_exc'], $summary['total_shipping']),
            'totalDiscounts' => self::price($summary['total_discounts_tax_exc'], $summary['total_discounts']),
            'totalWrapping' => self::price($summary['total_wrapping_tax_exc'], $summary['total_wrapping']),
            'totalPrice' => self::price($summary['total_price_without_tax'], $summary['total_price']),
            'totalTax' => self::formatPrice($summary['total_tax']),
            'minimalPurchase' => self::formatPrice($minimalPurchase),
        ];
        return $ret;
    }

    /**
     * @param Country $country
     * @param int $stateId
     * @param int $lang
     * @param string $vatNumber
     *
     * @return int
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getDummyAddress($country, $stateId, $lang, $vatNumber)
    {
        $countryId = (int)$country->id;
        $stateId = (int)$stateId;
        $sql = "
      SELECT MAX(id_address) as `id`
      FROM " . _DB_PREFIX_ . "address
      WHERE id_country = $countryId
        AND id_state = $stateId
        AND IFNULL(vat_number, '')='" . pSQL($vatNumber) . "'
        AND active = 1
        AND deleted = 0
        AND alias = 'chex'
      ";
        $result = Db::getInstance(_PS_USE_SQL_SLAVE_)->executeS($sql);
        if ($result && isset($result[0]['id'])) {
            return (int)$result[0]['id'];
        }
        $name = $country->name;

        if ($stateId) {
            $state = new State($stateId, $lang);
            if (!Validate::isLoadedObject($state)) {
                throw new PrestaShopException("Failed to load state $stateId");
            }
            $name .= ' / ' . $state->name;
        }

        $address = new Address();
        $address->id_country = $countryId;
        $address->id_state = $stateId;
        $address->alias = 'chex';
        $address->firstname = 'dummy';
        $address->lastname = 'address';
        $address->address1 = 'Dummy address created by chex module to calculate shipping rates to ' . $name;
        $address->vat_number = $vatNumber;
        $address->city = $name;
        $address->save();
        return (int)$address->id;
    }

    /**
     * @param float $value
     *
     * @return array
     * @throws PrestaShopException
     */
    private static function formatPrice($value)
    {
        return [
            'value' => $value,
            'formatted' => Tools::displayPrice($value)
        ];
    }

    /**
     * @param float $value
     * @param float $valueWithTax
     *
     * @return array
     * @throws PrestaShopException
     */
    private static function price($value, $valueWithTax)
    {
        return [
            'tax' => self::formatPrice($valueWithTax),
            'noTax' => self::formatPrice($value)
        ];
    }

    /**
     * @param Country $country
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getDeliveryOptionList($country)
    {
        $cart = $this->context->cart;
        Cache::clean('getPackageShippingCost*');
        Cache::clean('Carrier::getAvailableCarrierList*');
        $options = $cart->getDeliveryOptionList($country, true);
        $freeShipping = $this->freeShippingCartRuleExists($cart);
        $ret = [];
        if (count($options) == 1) {
            foreach ($options as $arr) {
                foreach ($arr as $key => $option) {
                    // at the moment, we don't support multiple carriers option
                    if (count($option['carrier_list']) === 1) {
                        $carrier = array_keys($option['carrier_list'])[0];
                        $cost = ($freeShipping || $option['is_free'])
                            ? self::price(0.0, 0.0)
                            : self::price($option['total_price_without_tax'], $option['total_price_with_tax']);
                        $ret[$key] = [
                            'carrier' => $carrier,
                            'cost' => $cost
                        ];
                    }
                }
            }
        }
        return $ret;
    }

    /**
     * @param Cart $cart
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function freeShippingCartRuleExists($cart)
    {
        foreach ($cart->getCartRules() as $rule) {
            if ($rule['free_shipping'] && !$rule['carrier_restriction']) {
                return true;
            }
        }
        return false;
    }

    /**
     * @return mixed
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getData()
    {
        return $this->withContext(function ($context) {
            $cart = $this->context->cart;
            $country = new Country($context['country']);
            $deliveryOptions = $this->getDeliveryOptionList($country);
            $delivery = $context['delivery'];
            $deliveryAddress = $context['deliveryAddress'];
            if (!$delivery || !isset($deliveryOptions[$delivery])) {
                $deliveryOption = $cart->getDeliveryOption($country, false, true);
                if (isset($deliveryOption[$deliveryAddress])) {
                    $delivery = $deliveryOption[$deliveryAddress];
                    if (!isset($deliveryOptions[$delivery])) {
                        $delivery = '';
                    }
                }

                if (! $delivery) {
                    $keys = array_keys($deliveryOptions);
                    if (count($keys)) {
                        $delivery = $keys[0];
                    }
                }
            }
            if ($delivery && isset($deliveryOptions[$delivery])) {
                $cart->id_carrier = $deliveryOptions[$delivery]['carrier'];
            }

            CartRule::autoRemoveFromCart($this->context);
            CartRule::autoAddToCart($this->context);

            $paymentOptions = $this->paymentsManager->getPaymentOptions();

            return $this->adjustData([
                'customer' => $this->customerManager->getCustomerData(),
                'cart' => $this->getCart(),
                'shipping' => [
                    'state' => $context['state'],
                    'country' => $context['country'],
                    'options' => $deliveryOptions,
                    'selected' => $delivery
                ],
                'payment' => [
                    'options' => $paymentOptions
                ],
                'deliveryAddress' => $this->checkAddress($context['deliveryAddress'], $this->customerManager->getAddresses(), $context['country'], $context['state']),
                'invoiceAddress' => $this->checkAddress($context['invoiceAddress'], $this->customerManager->getAddresses()),
            ]);
        }, $this->getContext());
    }

    /**
     * Returns first state in the country
     *
     * @param int $countryId
     * @return int
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getFirstState($countryId)
    {
        $res = State::getStatesByIdCountry($countryId);
        if ($res && isset($res[0]['id_state'])) {
            return (int)$res[0]['id_state'];
        }
        return 0;
    }

    /**
     * @param int $id
     * @param array $addresses
     * @param int|null $countryId
     * @param int|null $stateId
     * @return int|null
     */
    private function checkAddress($id, $addresses, $countryId = null, $stateId = null)
    {
        if ($id) {
            $id = (int)$id;
            $countryId = (int)$countryId;
            $stateId = (int)$stateId;
            foreach ($addresses as $address) {
                if ($address['id'] === $id) {
                    // check country
                    if ($countryId && $address['country'] !== $countryId) {
                        return null;
                    }
                    // check state
                    if ($stateId && $address['state'] !== $stateId) {
                        return null;
                    }
                    return $id;
                }
            }
        }
        return null;
    }


    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getContext()
    {
        $country = $this->context->country;
        $cookie = $this->context->cookie;
        $state = 0;
        if ($country->contains_states) {
            $state = isset($cookie->id_state)
              ? (int) $cookie->id_state
              : $this->getFirstState($country->id);
        }
        $context = [
            'country' => (int)$country->id,
            'state' => $state,
            'delivery' => '',
            'deliveryAddress' => null,
            'invoiceAddress' => null,
            'vatNumber' => '',
        ];
        $c = Tools::getValue('context');
        if ($c && is_array($c)) {
            if (isset($c['country'])) {
                $context['country'] = (int)$c['country'];
            }
            if (isset($c['state'])) {
                $context['state'] = (int)$c['state'];
            }
            if (isset($c['delivery'])) {
                $context['delivery'] = $c['delivery'];
            }
            if (isset($c['deliveryAddress']) && $c['deliveryAddress']) {
                $context['deliveryAddress'] = (int)$c['deliveryAddress'];
            }
            if (isset($c['invoiceAddress']) && $c['invoiceAddress']) {
                $context['invoiceAddress'] = (int)$c['invoiceAddress'];
            }
            if (isset($c['vatNumber']) && $c['vatNumber']) {
                $context['vatNumber'] = $c['vatNumber'];
            }
        }
        return $context;
    }

    /**
     * @param array $customerAddresses
     * @param Country $country
     * @param int $stateId
     * @param int $lang
     * @param string $vatNumber
     *
     * @return int
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getAddress($customerAddresses, $country, $stateId, $lang, $vatNumber)
    {
        $countryId = (int)$country->id;
        foreach ($customerAddresses as $addr) {
            if ($addr['country'] === $countryId && $addr['state'] === $stateId) {
                if ($vatNumber) {
                    if ($vatNumber === $addr['vatNumber']) {
                        return (int)$addr['id'];
                    }
                } else {
                    return (int)$addr['id'];
                }
            }
        }
        return $this->getDummyAddress($country, $stateId, $lang, $vatNumber);
    }

    /**
     * @param callable $func
     * @param array $context
     *
     * @return mixed
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     *
     * @noinspection PhpUndefinedFieldInspection
     */
    private function withContext($func, $context)
    {
        $cookie = $this->context->cookie;
        $countryId = (int)$context['country'];
        $stateId = (int)$context['state'];
        $vatNumber = $context['vatNumber'];
        $lang = (int)Configuration::get('PS_LANG_DEFAULT');
        $country = new Country($countryId, $lang);

        $delivery = $context['delivery'] ?? '';
        $context['delivery'] = $delivery;

        // determine delivery address
        $addresses = $this->customerManager->getAddresses();
        $deliveryAddress = $context['deliveryAddress'];
        if (!$this->checkAddress($deliveryAddress, $addresses, $countryId, $stateId)) {
            $deliveryAddress = $this->getAddress($addresses, $country, $stateId, $lang, $vatNumber);
            $context['deliveryAddress'] = $deliveryAddress;
        }

        // determine invoice address
        $invoiceAddress = $context['invoiceAddress'];
        if (!$this->checkAddress($invoiceAddress, $addresses)) {
            $invoiceAddress = $deliveryAddress;
            $context['invoiceAddress'] = $invoiceAddress;
        }

        $cart = $this->context->cart;
        $cartId = (int)$cart->id;

        // prepare up environment
        $save = [
            'delivery_option' => $cart->delivery_option,
            'id_address_delivery' => $cart->id_address_delivery,
            'id_address_invoice' => $cart->id_address_invoice
        ];
        $cart->delivery_option = $this->serializeDeliveryOption([$deliveryAddress => $delivery]);
        $cart->id_address_delivery = $deliveryAddress;
        $cart->id_address_invoice = $invoiceAddress;
        $cart->save();

        $this->context->country = $country;

        Db::getInstance()->update('cart_product', ['id_address_delivery' => $deliveryAddress], '`id_cart` = ' . $cartId);
        Db::getInstance()->update('customization', ['id_address_delivery' => $deliveryAddress], '`id_cart` = ' . $cartId);

        $cookie->id_country = $countryId;
        if ($stateId) {
          $cookie->id_state = $stateId;
        } else {
          unset($cookie->id_state);
        }

        try {
            return call_user_func($func, $context);
        } finally {
            $cart->delivery_option = $save['delivery_option'];
            $cart->id_address_delivery = $save['id_address_delivery'];
            $cart->id_address_invoice = $save['id_address_invoice'];
            $cart->save();
            Db::getInstance()->update('cart_product', ['id_address_delivery' => $cart->id_address_delivery], '`id_cart` = ' . $cartId);
            Db::getInstance()->update('customization', ['id_address_delivery' => $cart->id_address_delivery], '`id_cart` = ' . $cartId);
        }
    }

    /**
     * @param array $option
     * @return string
     */
    private function serializeDeliveryOption($option)
    {
        if (defined('_TB_VERSION_') && version_compare(_TB_VERSION_, '1.0.4', '>=')) {
            return json_encode($option);
        }
        return serialize($option);
    }

    /**
     * return Step Flow
     *
     * @return array
     * @throws PrestaShopException
     */
    private function getFlow()
    {
        switch ($this->module->getSettings()->getStepsFlow()) {
            case 'unified':
                return [ 'basic', 'shipping-options', 'shipping', 'billing', 'confirmation' ];
            case 'separate':
                return [ 'choose-shipping', 'shipping-options', 'choose-payment', 'shipping', 'billing', 'confirmation' ];
            default:
                throw new PrestaShopException('Invalid flow settings');
        }
    }

    /**
     * @return EntryFields
     */
    public function getEntryFieldsManager()
    {
        return $this->fieldsManager;
    }

    /**
     * @param string $message
     * @return string
     */
    private function l($message) {
        return $this->module->l($message);
    }

    /**
     * @param array $data
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function adjustData(array $data)
    {
        $modules = Hook::getHookModuleExecList('actionAdjustChexData');
        if ($modules && is_array($modules)) {
            foreach ($modules as $m) {
                $moduleId = (int) $m['id_module'];
                $ret = Hook::exec('actionAdjustChexData', [ 'chexData' => $data ], $moduleId, true);
                if ($ret && is_array($ret) && count($ret) === 1) {
                    $ret = array_values($ret);
                    $data = $ret[0];
                }
            }
        }
        return $data;
    }

    /**
     * @param int $deliveryAddressId
     * @param int $invoiceAddressId
     * @param string $deliveryOption
     *
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function setAddresses($deliveryAddressId, $invoiceAddressId, $deliveryOption)
    {
        $deliveryAddressId = (int)$deliveryAddressId;
        $invoiceAddressId = (int)$invoiceAddressId;

        $addresses = array_map('intval', array_column($this->customerManager->getAddresses(), 'id'));

        if (! in_array($deliveryAddressId, $addresses)) {
            throw new PrestaShopException("Invalid delivery address id");
        }

        if (! in_array($invoiceAddressId, $addresses)) {
            throw new PrestaShopException("Invalid invoice address id");
        }

        $cart = Context::getContext()->cart;
        $cart->id_address_delivery = $deliveryAddressId;
        $cart->id_address_invoice = $invoiceAddressId;
        $cart->delivery_option = $this->serializeDeliveryOption([$deliveryAddressId => $deliveryOption]);
        $cart->save();

        Db::getInstance()->update('cart_product', ['id_address_delivery' => $deliveryAddressId], '`id_cart` = ' . $cart->id);
        Db::getInstance()->update('customization', ['id_address_delivery' => $deliveryAddressId], '`id_cart` = ' . $cart->id);

        return true;
    }
}
