<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Hook;
use Module;
use PrestaShopDatabaseException;
use PrestaShopException;
use PrestaShopLogger;

class PaymentsManager
{
    /**
     * @return array
     * @throws PrestaShopException
     */
    public function getPaymentOptions()
    {
        return $this->getDisplayPaymentEUOptions();
    }

    /**
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function getDisplayPaymentEUOptions()
    {
        $ret = [];
        $paymentOptions = Hook::exec('displayPaymentEU', [], null, true);
        if (is_array($paymentOptions)) {
            $options = [];
            foreach ($paymentOptions as $module => $arr) {
                if (Module::isEnabled($module) && is_array($arr)) {
                    if (array_key_exists('cta_text', $arr)) {
                        $arr = [$arr];
                    }
                    foreach ($arr as $key => $option) {
                        if (array_key_exists('cta_text', $option)) {
                            $option = self::fromPaymentEUOption($module, $option);
                            if ($option) {
                                $id = $option['id'];
                                if (isset($options[$id])) {
                                    $id = md5($key . $id);
                                    $option['id'] = $id;
                                }
                                $options[$id] = $option;
                            }
                        }
                    }
                }
            }
            $ret = array_values($options);
        }
        return $ret;
    }

    /**
     * @param string $module
     * @param array $option
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private static function fromPaymentEUOption($module, $option)
    {
        $action = static::getPaymentEUAction($option);
        if ($action) {
            $ret = [
                'id' => static::getPaymentOptionId($module, $option),
                'module' => $module,
                'label' => $option['cta_text'],
                'action' => $action
            ];
            if (isset($option['logo']) && is_string($option['logo'])) {
                $ret['logo'] = $option['logo'];
            }
            if (isset($option['additionalInformation'])) {
                $ret['additionalInformation'] = $option['additionalInformation'];
            }
            return $ret;
        } else {
            PrestaShopLogger::addLog("Chex: Failed to convert payment option " . $option['cta_text'], 3);
            return null;
        }
    }

    /**
     * @param array $option
     * @return array|null
     */
    private static function getPaymentEUAction($option)
    {
        // try redirect binding
        if (isset($option['action'])) {
            $action = $option['action'];
            if (is_string($action) && $action) {
                return [
                    'type' => 'redirect',
                    'url' => $action
                ];
            }
        }

        // try form binding
        if (isset($option['form'])) {
            $form = $option['form'];
            if (is_string($form)) {
                $form = str_replace("@hiddenSubmit", "", $form);
                if ($form) {
                    return [
                        'type' => 'form',
                        'form' => $form
                    ];
                }
            }
        }
        return null;
    }

    /**
     * @param string $module
     * @param array $option
     * @return string
     */
    private static function getPaymentOptionId($module, $option)
    {
        if (isset($option['action'])) {
            $action = $option['action'];

            $url = urldecode(trim($action));
            $parsedUrl = parse_url($url);
            if (isset($parsedUrl['query'])) {
                $queryParts = [];
                parse_str($parsedUrl['query'], $queryParts);
                foreach (static::getIgnoredQueryKeys() as $ignore) {
                    unset($queryParts[$ignore]);
                }
                ksort($queryParts);
                $parsedUrl['query'] = http_build_query($queryParts);
                $url = http_build_url('', $parsedUrl);
            }
            return md5($module . ':' . $url);
        }

        return md5($module . ':' . $option['cta_text']);
    }

    /**
     * Returns list of ignored parameters from payment option id
     * @return array
     */
    private static function getIgnoredQueryKeys()
    {
        return [
            'rand',
            'rnd',
            'time',
            'date',
            'ts',
        ];
    }

}
