<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Configuration;
use Chex;
use ImageType;
use Module;
use PrestaShopDatabaseException;
use PrestaShopException;

class Settings
{
    const LICENSE = 'CHEX_LICENSE';
    const APP_URL = 'CHEX_APP_URL';
    const BACKEND_APP_URL = 'CHEX_BACK_APP_URL';
    const API_URL = 'CHEX_API_URL';
    const SETTINGS = 'CHEX_SETTINGS';
    const VERSION = 'CHEX_VERSION';
    const ACTIVATED = 'CHEX_ACTIVE';
    const CHECK_VERSION = 'CHEX_CHECK_VERSION';

    /** @var array */
    private $data;

    /**
     * Settings constructor.
     *
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function __construct()
    {
        $this->data = self::getDefaultSettings();
        $stored = Configuration::get(self::SETTINGS);
        if ($stored) {
            $stored = json_decode($stored, true);
            if ($stored) {
                $this->data = self::mergeSettings($this->data, $stored);
            }
            if (! $this->data['display']['imageType']) {
                $this->data['display']['imageType'] = static::getBestImageType();
                $this->set($this->data);
            }
        }
    }

    /**
     * @return array
     */
    private static function getDefaultSettings()
    {
        return [
            'sandbox' => [
                'enabled' => true,
                'addresses' => []
            ],
            'layout' => [
                'cart' => 'left',
                'floatingCart' => true,
                'flow' => 'unified',
            ],
            'display' => [
                'carrierLogo' => true,
                'paymentLogo' => true,
                'imageType' => '',
                'countrySelector' => 'all',
            ]
        ];
    }

    /**
     * @param mixed $val
     * @return bool
     */
    private static function toBool($val)
    {
        return !!$val;
    }

    /**
     * @return bool
     * @throws PrestaShopException
     */
    public function init()
    {
        return $this->set(self::getDefaultSettings());
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function reset()
    {
        return $this->remove() && $this->init();
    }

    /**
     * @return string
     * @throws PrestaShopException
     */
    public function getLicense()
    {
        return Configuration::get(self::LICENSE);
    }

    /**
     * @param string $license
     *
     * @return bool
     * @throws PrestaShopException
     */
    public function setLicense($license)
    {
        return Configuration::updateValue(self::LICENSE, $license);
    }


    /**
     * @return bool
     * @throws PrestaShopException
     */
    public function isActivated()
    {
        return (bool)Configuration::get(self::ACTIVATED);
    }

    /**
     * @return string
     */
    public function getCartPosition()
    {
        $pos = $this->get(['layout', 'cart']);
        if (in_array($pos, ['left', 'right', 'top'])) {
            return $pos;
        }
        return 'left';
    }

    /**
     * @return string
     */
    public function getStepsFlow()
    {
        $flow = $this->get(['layout', 'flow']);
        if (in_array($flow, ['unified', 'separate'])) {
            return $flow;
        }
        return 'unified';
    }

    /**
     * @return bool
     */
    public function showCarrierLogo()
    {
        return static::toBool($this->get(['display', 'carrierLogo']));
    }

    /**
     * @return bool
     */
    public function showPaymentLogo()
    {
        return static::toBool($this->get(['display', 'paymentLogo']));
    }

    /**
     * @return bool
     */
    public function allowFloatingCart()
    {
        return static::toBool($this->get(['layout', 'floatingCart']));
    }

    /**
     * @return bool
     * @throws PrestaShopException
     */
    public function setActivated()
    {
        return Configuration::updateValue(self::ACTIVATED, 1);
    }

    /**
     * @param Chex $module
     * @return string
     * @throws PrestaShopException
     */
    public function getAppUrl($module)
    {
        $url = Configuration::get(self::APP_URL);
        if (!$url) {
            $version = self::getUnderscoredVersion($module);
            $url = $module->getPath("/views/js/chex-$version.js");
        }
        return $url;
    }

    /**
     * @param Chex $module
     * @return string
     * @throws PrestaShopException
     */
    public function getBackendAppUrl($module)
    {
        $url = Configuration::get(self::BACKEND_APP_URL);
        if (!$url) {
            $version = self::getUnderscoredVersion($module);
            $url = $module->getPath("views/js/back-$version.js");
        }
        return $url;
    }

    /**
     * @return string
     * @throws PrestaShopException
     */
    public function getApiUrl()
    {
        return Configuration::get(self::API_URL);
    }

    /**
     * @return string
     * @throws PrestaShopException
     */
    public function getVersion()
    {
        return Configuration::get(self::VERSION);
    }

    /**
     * @param string $version
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function setVersion($version)
    {
        Configuration::deleteByName(self::CHECK_VERSION);
        Configuration::updateValue(self::VERSION, $version);
    }

    /**
     * @return array
     * @throws PrestaShopException
     */
    public function getCheckModuleVersion()
    {
        $ret = Configuration::get(self::CHECK_VERSION);
        if ($ret) {
            return json_decode($ret, true);
        }
        return ['ts' => null, 'version' => null, 'notes' => null];
    }

    /**
     * @param string $version
     * @param int $ts
     * @param string $notes
     * @throws PrestaShopException
     */
    public function setCheckModuleVersion($version, $ts, $notes)
    {
        Configuration::updateValue(self::CHECK_VERSION, json_encode([
            'ts' => $ts,
            'version' => $version,
            'notes' => $notes,
        ]));
    }

    /**
     * @return bool
     */
    public function isSandboxEnabled()
    {
        return (bool)$this->get(['sandbox', 'enabled']);
    }

    /**
     * @return array|mixed
     */
    public function getSandboxAddresses()
    {
        return $this->get(['sandbox', 'addresses']);
    }

    /**
     * @return string
     */
    public function getImageType()
    {
        return $this->get(['display', 'imageType']);
    }

    /**
     * @return string - 'all' | 'active'
     */
    public function getCountrySelectorMode()
    {
        if ($this->get(['display', 'countrySelector']) === 'all') {
            return 'all';
        }
        return 'active';
    }

    /**
     * @return string
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private static function getBestImageType()
    {
        $best = null;
        foreach (ImageType::getImagesTypes('products', true) as $imageType) {
            $width = (int)$imageType['width'];
            $height = (int)$imageType['height'];
            if ($width >= 70 && $height >= 70) {
                $best = $imageType['name'];
            }
        }
        return $best;
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function remove()
    {
        $this->data = null;
        Configuration::deleteByName(self::CHECK_VERSION);
        Configuration::deleteByName(self::SETTINGS);
        return true;
    }

    /**
     * @param null $path
     * @return array|mixed
     */
    public function get($path = null)
    {
        $value = $this->data;
        if (is_null($path)) {
            return $value;
        }
        foreach ($path as $key) {
            if (isset($value[$key])) {
                $value = $value[$key];
            } else {
                die('Chex: setting not found: ' . implode($path, '>'));
            }
        }
        return $value;
    }

    /**
     * @param mixed $value
     * @return bool
     * @throws PrestaShopException
     */
    public function set($value)
    {
        $this->data = $value;
        return Configuration::updateValue(self::SETTINGS, json_encode($value));
    }

    /**
     * @param array $left
     * @param array $right
     * @return array
     */
    private static function mergeSettings($left, $right)
    {
        $ret = [];
        foreach ($left as $key => $value) {
            if (isset($right[$key])) {
                if (is_array($value) && self::isAssoc($value)) {
                    $value = self::mergeSettings($value, $right[$key]);
                } else {
                    $value = $right[$key];
                }
            }
            $ret[$key] = $value;
        }
        return $ret;
    }

    /**
     * @param array $arr
     * @return bool
     */
    private static function isAssoc(array $arr)
    {
        if (array() === $arr) {
            return false;
        }
        return array_keys($arr) !== range(0, count($arr) - 1);
    }

    /**
     * @param Module $module
     *
     * @return string
     */
    private static function getUnderscoredVersion($module)
    {
        return str_replace('.', '_', $module->version);
    }

}
