<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

class AdminChexBackendController extends ModuleAdminController
{
    /** @var Chex */
    public $module;

    /**
     * AdminChexBackendController constructor.
     * @throws PrestaShopException
     */
    public function __construct()
    {
        parent::__construct();
        $this->display = 'view';
        $this->bootstrap = false;
        $this->addCSS($this->getPath('views/css/back-0_9_3.css'));
        $this->addCSS($this->getPath('views/css/chex-0_9_3.css'));
        $this->addJs($this->module->getSettings()->getBackendAppUrl($this->module));
    }

    /**
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     * @throws SmartyException
     */
    public function display()
    {
        $settings = $this->module->getSettings();
        $this->display_footer = false;
        $platform = 'prestashop';
        $platformVersion = _PS_VERSION_;
        if (defined('_TB_VERSION_')) {
            $platform = 'thirtybees';
            $platformVersion = _TB_VERSION_;
        }

        $entryFieldsManager = $this->module->getManager()->getEntryFieldsManager();

        $chex = [
            'ip' => Tools::getRemoteAddr(),
            'apiUrl' => $this->context->link->getAdminLink('AdminChexBackend'),
            'activated' => $settings->isActivated(),
            'version' => $this->module->version,
            'platform' => $platform,
            'platformVersion' => $platformVersion,
            'versionCheck' => $settings->getCheckModuleVersion(),
            'settings' => $settings->get(),
            'license' => $settings->getLicense(),
            'translations' => $this->module->getBackTranslations(),
            'fields' => $entryFieldsManager->getFields(true),
            'fieldPreferences' => $entryFieldsManager->getFieldsPreferences(),
            'imageTypes' => $this->getImageTypes()
        ];

        $storeUrl = $settings->getApiUrl();
        if ($storeUrl) {
            $chex['storeUrl'] = $storeUrl;
        }

        $this->context->smarty->assign([
            'help_link' => null,
            'title' => $this->l('Checkout Experience'),
            'chex' => $chex
        ]);

        parent::display();
    }

    /**
     * @param string $tpl_name
     *
     * @return object|Smarty_Internal_Template
     * @throws PrestaShopException
     * @throws SmartyException
     */
    public function createTemplate($tpl_name)
    {
        if ($this->viewAccess() && $tpl_name === 'content.tpl') {
            $path = _PS_MODULE_DIR_ . $this->module->name . '/views/templates/admin/backend.tpl';
            return $this->context->smarty->createTemplate($path, $this->context->smarty);
        }
        return parent::createTemplate($tpl_name);
    }

    /**
     * @param string $path
     * @return string
     */
    private function getPath($path)
    {
        return $this->module->getPath($path);
    }

    /**
     *
     */
    public function ajaxProcessCommand()
    {
        $error = null;
        $result = null;
        try {
            $result = $this->dispatchCommand(Tools::getValue('cmd'));
        } catch (Exception $e) {
            $error = $e->getMessage();
        }
        $this->reply($error, $result);
        die();
    }

    /**
     * @param string $cmd
     *
     * @return array|bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function dispatchCommand($cmd)
    {
        $rawPayload = $_POST['payload'] ?? '';
        $payload = json_decode($rawPayload, true);
        switch ($cmd) {
            case 'setLatestVersion':
                return $this->setLatestVersion($payload);
            case 'setLicense':
                return $this->setLicense($payload);
            case 'setSettings':
                return $this->setSettings($payload);
            case 'accountActivated':
                return $this->accountActivated();
            case 'enableField':
                return $this->enableField($payload);
            case 'saveField':
                return $this->saveField($payload);
            case 'deleteField':
                return $this->deleteField($payload);
            default:
                throw new PrestaShopException("Unknown command $cmd");
        }
    }


    /**
     * @param array $data
     *
     * @return bool
     * @throws PrestaShopException
     */
    private function setLatestVersion($data)
    {
        if (!isset($data['version'])) {
            throw new PrestaShopException('Version not set');
        }
        if (!isset($data['ts'])) {
            throw new PrestaShopException('Timestamp not set');
        }
        $this->module->getSettings()->setCheckModuleVersion($data['version'], $data['ts'], $data['notes']);
        return true;
    }

    /**
     * @param array $data
     * @return bool
     * @throws PrestaShopException
     */
    private function setLicense($data)
    {
        if (!isset($data['license'])) {
            throw new PrestaShopException('License not set');
        }
        $this->module->getSettings()->setLicense($data['license']);
        return true;
    }

    /**
     * @param array $data
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function saveField($data)
    {
        if (!isset($data['field'])) {
            throw new PrestaShopException('Field not set');
        }
        $field = $data['field'];
        if (! isset($field['id'])) {
            throw new PrestaShopException('Field id not set');
        }
        $id = $field['id'];
        $entryFieldManager = $this->module->getManager()->getEntryFieldsManager();
        $entryFieldManager->saveField($field);
        return [
            'field' => $entryFieldManager->getField($id),
            'preferences' => $entryFieldManager->getFieldPreferences($id)
        ];
    }

    /**
     * @param array $data
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function deleteField($data)
    {
        if (!isset($data['id'])) {
            throw new PrestaShopException('Id not set');
        }
        $id = $data['id'];
        $entryFieldManager = $this->module->getManager()->getEntryFieldsManager();
        return $entryFieldManager->deleteField($id);
    }


    /**
     * @param array $data
     * @return array
     * @throws PrestaShopException
     */
    private function enableField($data)
    {
        if (!isset($data['field'])) {
            throw new PrestaShopException('field not set');
        }
        if (!isset($data['enable'])) {
            throw new PrestaShopException('enable not set');
        }
        $field = $data['field'];
        $entryFieldManager = $this->module->getManager()->getEntryFieldsManager();
        $entryFieldManager->enableField($field, !!$data['enable']);
        return $entryFieldManager->getFieldPreferences($field);
    }


    /**
     * @param array $data
     * @return bool
     * @throws PrestaShopException
     */
    private function setSettings($data)
    {
        if (!isset($data['settings'])) {
            throw new PrestaShopException('Settings not set');
        }
        $this->module->getSettings()->set($data['settings']);
        return true;
    }

    /**
     * @return bool
     * @throws PrestaShopException
     */
    private function accountActivated()
    {
        $this->module->getSettings()->setActivated();
        return true;
    }

    /**
     * @param string $error
     * @param mixed $result
     */
    private function reply($error, $result)
    {
        if ($error) {
            echo json_encode(['success' => false, 'error' => $error]);
        } else {
            echo json_encode(['success' => true, 'result' => $result]);
        }
    }

    /**
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function getImageTypes()
    {
        $types = ImageType::getImagesTypes('products', true);
        return array_reverse(array_map(function($type) {
            return [
                'name' => $type['name'],
                'dimensions' => $type['width'] . 'x' . $type['height']
            ];
        }, $types));
    }

}
