<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

use Chex\EntryFields;

class ChexCheckoutModuleFrontController extends ModuleFrontController
{
    /**
     * ChexCheckoutModuleFrontController constructor.
     *
     * @throws PrestaShopException
     */
    public function __construct()
    {
        parent::__construct();
        $this->php_self = 'module-chex-checkout';
        $this->display_column_left = false;
        $this->display_column_right = false;
    }

    /**
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function initContent()
    {
        if (!$this->getModule()->useChex()) {
            $type = Configuration::get('PS_ORDER_PROCESS_TYPE') ? 'order-opc' : 'order';
            Tools::redirect($this->context->link->getPageLink($type, true));
        }
        parent::initContent();
        if (Tools::isSubmit('action') && Tools::getValue('action') == 'command') {
            $this->ajaxProcessCommand();
        } else {
            $module = $this->getModule();
            $this->addCSS(_PS_ROOT_DIR_ . '/modules/chex/views/css/chex-0_9_3.css');
            $this->addJS(_PS_JS_DIR_ . 'validate.js');
            $this->addJS($module->getSettings()->getAppUrl($module));
            $this->context->smarty->assign([
                'chexData' => $module->getManager()->getInitialData(),
                'body_classes' => [strtolower('chex-' . _THEME_NAME_)]
            ]);
            $this->setTemplate('checkout.tpl');
        }
    }

    /**
     *
     */
    public function ajaxProcessCommand()
    {
        try {
            $this->getModule()->csrf()->validate(self::getValue('chexToken'));
            $result = $this->dispatchCommand(self::getValue('cmd'));
            die(json_encode(['success' => true, 'result' => $result]));
        } catch (Exception $e) {
            $error = $e->getMessage();
            die(json_encode(['success' => false, 'error' => $error]));
        }
    }

    /**
     * @param string $cmd
     * @return mixed
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function dispatchCommand($cmd)
    {
        $manager = $this->getModule()->getManager();
        $customerManager = $manager->getCustomerManager();
        switch ($cmd) {
            case 'refreshData':
                return $manager->getData();
            case 'addCoupon':
                $this->addCoupon(self::getValue('coupon'));
                return $manager->getData();
            case 'checkEmail':
                return $customerManager->customerExists(self::getValue('email', array('Validate', 'isEmail'))) ? 'customer' : 'visitor';
            case 'checkVatNumber':
                return $manager->checkVatNumber(self::getValue('vatNumber'));
            case 'checkAvailability':
                return $manager->checkAvailability();
            case 'createAddress':
                return $customerManager->createAddress(self::getValue('address', 'is_array'));
            case 'setAddress':
                return $manager->setAddresses(
                    self::getValue('deliveryAddressId', ['Validate', 'isInt']),
                    self::getValue('invoiceAddressId', ['Validate', 'isInt']),
                    self::getValue('deliveryOption')
                );
            case 'createCustomer':
                return $customerManager->createCustomer(self::getValue('customer', 'is_array'));
            case 'processCustomData':
                return $this->processCustomData(self::getValue('data', 'is_string'));
            case 'processMessage':
                return $this->processMessage(self::getValue('message', 'is_string'));
            case 'getSelectOptions':
                return $this->getSelectOptions(self::getValue('search', 'is_string'), self::getValue('description', 'is_array'));
            case 'resetPassword':
                return $this->resetPassword(self::getValue('email', 'is_string'));
            case 'editCart':
                return $this->editCart(self::getValue('products', 'is_array'));
            case 'createFreeOrder':
                return $this->createFreeOrder();
            default:
                throw new PrestaShopException("Unknown command $cmd");
        }
    }

    /**
     * @param string $search
     * @param array $description
     *
     * @return array
     * @throws PrestaShopException
     */
    private function getSelectOptions($search, $description)
    {
        switch ($description['type']) {
            case 'carrier':
                $carrierManager = $this->getModule()->getManager()->getCarrierManager();
                $ret = $carrierManager->getSelectOptions((int)$description['carrierId'], $description['field'], $search);
                break;
            case 'entryField':
                $integration = $this->getModule()->getIntegrations()->getEntryFieldsIntegration($description['integration']);
                if (! $integration) {
                    throw new PrestaShopException('Integration not found');
                }
                $ret = $integration->getSelectOptions($description['field'], $search);
                break;
            default:
                throw new PrestaShopException('Invalid type ' . $description['type']);
        }
        return [
            'search' => $search,
            'description' => $description,
            'result' => $ret
        ];
    }

    /**
     * @param string $data
     *
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function processCustomData($data)
    {
        $data = json_decode($data, true);
        $this->saveCustomData($data);

        if ($data && is_array($data)) {
            foreach ($data as $section) {
                if ($section['type'] === 'shipping') {
                    $carrierManager = $this->getModule()->getManager()->getCarrierManager();
                    $carrierManager->processCustomData((int)$section['carrierId'], $section['form']);
                }
                if ($section['type'] === 'address') {
                    $address = new Address((int)$section['addressId']);
                    $this->dispatchCustomData('address', $section['form'], $address);
                }
                if ($section['type'] === 'confirm') {
                    $cart = $this->context->cart;
                    $this->dispatchCustomData('confirm', $section['form'], $cart);
                }
                if ($section['type'] === 'account') {
                    $customer = new Customer((int)$section['customerId']);
                    $this->dispatchCustomData('account', $section['form'], $customer);
                }
            }
        }
        return true;
    }

    /**
     * Saves custom field values to db
     *
     * @param array $data
     *
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function saveCustomData($data)
    {
        $values = [
            'chex_field_value_bool' => [],
            'chex_field_value_int' => [],
            'chex_field_value_text' => [],
            'chex_field_value_longtext' => [],
        ];

        $cartId = (int)$this->context->cart->id;
        if ($cartId && $data && is_array($data)) {
            $entryFields = $this->getModule()->getManager()->getEntryFieldsManager();
            foreach ($data as $section) {
                $sectionType = $section['type'];
                $customEntryFields = $entryFields->getCustomFields($sectionType);
                $form = $section['form'];
                $entityId = $cartId;
                if ($sectionType === 'shipping') {
                    $entityId = (int)$section['carrierId'];
                }
                if ($section['type'] === 'address') {
                    $entityId = (int)$section['addressId'];
                }
                if ($section['type'] === 'account') {
                    $entityId = (int)$section['customerId'];
                }
                foreach ($customEntryFields as $field) {
                    $fullId = $field['id'];
                    $parts = explode('.', $fullId);
                    $fieldId = $parts[1];
                    $value = $form[$fieldId] ?? $entryFields->getFieldDefaultValue($field);
                    switch ($field['type']) {
                        case EntryFields::TYPE_CHECKBOX:
                            $values['chex_field_value_bool'][] = $this->fieldValueInsert($cartId, $fullId, !!$value, $sectionType, $entityId);
                            break;
                        case EntryFields::TYPE_COUNTRY:
                        case EntryFields::TYPE_STATE:
                            $values['chex_field_value_int'][] = $this->fieldValueInsert($cartId, $fullId, (int)$value, $sectionType, $entityId);
                            break;
                        case EntryFields::TYPE_PASSWORD:
                            $values['chex_field_value_text'][] = $this->fieldValueInsert($cartId, $fullId, 'SHA1=' . sha1($value), $sectionType, $entityId);
                            break;
                        case EntryFields::TYPE_EMAIL:
                        case EntryFields::TYPE_TEL:
                        case EntryFields::TYPE_NUMBER:
                        case EntryFields::TYPE_TEXT:
                        case EntryFields::TYPE_SELECT:
                            $values['chex_field_value_text'][] = $this->fieldValueInsert($cartId, $fullId, $value, $sectionType, $entityId);
                            break;
                        case EntryFields::TYPE_TEXTAREA:
                            $values['chex_field_value_longtext'][] = $this->fieldValueInsert($cartId, $fullId, $value, $sectionType, $entityId);
                            break;
                        default:
                            throw new PrestaShopException("Invalid field type: " . $field['type']);
                    }
                }
            }
        }

        $conn = Db::getInstance();
        foreach ($values as $table => $insertValues) {
            foreach ($insertValues as $insertValue) {
                $conn->delete($table, "id_cart = $cartId AND field = '" .$insertValue['field']. "' and id_entity = " . $insertValue['id_entity']);
                $conn->insert($table, $insertValue);
            }
        }

    }

    /**
     * @param int $cartId
     * @param string $field
     * @param string $value
     * @param string $section
     * @param int $entityId
     *
     * @return array
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function fieldValueInsert($cartId, $field, $value, $section, $entityId)
    {
        return [
            'id_cart' => (int)$cartId,
            'field' => pSQL($field),
            'id_entity' =>(int)$entityId,
            'value' => pSQL($value, true),
            'section' => pSQL($section),
        ];
    }


    /**
     * @param string $messageContent
     *
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function processMessage($messageContent)
    {
        $cartId = (int)$this->context->cart->id;

        if ($cartId && $messageContent) {

            if (! Validate::isMessage($messageContent)) {
                throw new PrestaShopException('Invalid message');
            }

            $message = new Message();
            $message->message = $messageContent;
            $message->id_cart = $cartId;
            $message->id_customer = (int) $this->context->cart->id_customer;
            $message->add();
        }

        return true;
    }

    /**
     * @param string $section
     * @param array $form
     * @param ObjectModel $object
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function dispatchCustomData($section, $form, $object)
    {
        $entryFields = $this->getModule()->getManager()->getEntryFieldsManager();
        $customEntryFields = $entryFields->getCustomFields($section);
        $integrations = $this->getModule()->getIntegrations()->getEntryFieldsIntegrations();
        foreach ($customEntryFields as $field) {
            $custom = $field['custom'];
            $fullId = $field['id'];
            $parts = explode('.', $fullId);
            $fieldId = $parts[1];
            $value = $form[$fieldId] ?? $entryFields->getFieldDefaultValue($field);
            if (isset($integrations[$custom])) {
                $integration = $integrations[$custom];
                $integration->processEntryField($fullId, $value, $object);
            }
        }
    }


    /**
     * @param string $code
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    private function addCoupon($code)
    {
        $code = trim($code);
        $defaultError = 'Invalid coupon code';
        if (CartRule::isFeatureActive()) {
            if (!$code || !Validate::isCleanHtml($code)) {
                throw new PrestaShopException($defaultError);
            }
            $cartRule = new CartRule(CartRule::getIdByCode($code));
            if (!Validate::isLoadedObject($cartRule)) {
                throw new PrestaShopException($defaultError);
            }
            $error = $cartRule->checkValidity($this->context, false, true);
            if ($error) {
                throw new PrestaShopException($error);
            }
            $this->context->cart->addCartRule($cartRule->id);
            CartRule::autoAddToCart($this->context);
        } else {
            throw new PrestaShopException($defaultError);
        }
    }

    /**
     * @param string $email
     * @return boolean
     * @throws PrestaShopException
     */
    private function resetPassword($email)
    {
        $customerManager = $this->getModule()->getManager()->getCustomerManager();
        return $customerManager->resetPasswordRequest($email);
    }

    /**
     * @param array $products
     * @return array
     * @throws PrestaShopException
     */
    private function editCart($products)
    {
        $cart = Context::getContext()->cart;
        foreach ($products as $product) {
            if (! isset($product['productId']) || !(int)$product['productId']) {
                throw new PrestaShopException('productId missing');
            }
            $productId = (int)$product['productId'];

            if (! isset($product['combinationId'])) {
                throw new PrestaShopException('combinationId missing');
            }
            $combinationId = (int)$product['combinationId'];

            if (! isset($product['diff'])) {
                throw new PrestaShopException('diff missing');
            }
            $diff = (int)$product['diff'];
            $dir = $diff > 0 ? 'up' : 'down';

            $cart->updateQty(abs($diff), $productId, $combinationId ? $combinationId : null, false, $dir);
        }
        return $this->getModule()->getManager()->getData();
    }

    /**
     * @param string $name
     * @param callable|null $validator
     * @return mixed
     * @throws PrestaShopException
     */
    private static function getValue($name, $validator = null)
    {
        if (! isset($_POST[$name])) {
            throw new PrestaShopException("Missing parameter: $name");
        }
        $value = $_POST[$name];
        if ($validator) {
            if (!call_user_func($validator, $value)) {
                throw new PrestaShopException("Invalid parameter: $name");
            }
        }
        return $value;
    }

    /**
     * @return Chex
     * @throws PrestaShopException
     */
    private function getModule()
    {
        if ($this->module instanceof Chex) {
            return $this->module;
        }
        throw new PrestaShopException('Failed to resolve module');
    }

    /**
     * @return array
     *
     * @throws PrestaShopException
     * @throws SmartyException
     */
    private function createFreeOrder()
    {
        $cart = $this->context->cart;
        if ($cart->getOrderTotal() > 0) {
            throw new PrestaShopException("Can't create free order");
        }

        $freePaymentMethod = new FreeOrder();
        if (! $freePaymentMethod->validateOrder(
            $cart->id,
            Configuration::get('PS_OS_PAYMENT'),
            0,
            Tools::displayError('Free order', false),
            null,
            [],
            null,
            false,
            $cart->secure_key
        )) {
            throw new PrestaShopException("Can't validate free order");
        };

        $order = new Order((int)$freePaymentMethod->currentOrder);
        if (! Validate::isLoadedObject($order)) {
            throw new PrestaShopException("Failed to create free order");
        }

        $customer = $this->context->customer;
        $link = $this->context->link;
        if ($customer->is_guest) {
            $email = $customer->email;
            $customer->mylogout();
            $redirect = $link->getPageLink('guest-tracking', null, null, 'id_order='.urlencode($order->reference).'&email='.urlencode($email));
        } else {
            $redirect = $this->context->link->getPageLink('history');
        }

        return [
            'orderId' => (int)$freePaymentMethod->currentOrder,
            'redirect' => $redirect,
        ];

    }

}
