<?php
/**
 * Copyright (C) 2017-2019 Petr Hucik <petr@getdatakick.com>
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@getdatakick.com so we can send you a copy immediately.
 *
 * @author    Petr Hucik <petr@getdatakick.com>
 * @copyright 2017-2019 Petr Hucik
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

namespace Chex;

use Cart;
use Configuration;
use Db;
use DbQuery;
use PrestaShopDatabaseException;
use PrestaShopException;
use Translate;

class ShaimZasilkovnaAdapter implements ShippingIntegration
{
    /**
     * @param array $carrier
     *
     * @return array
     */
    function getCarrierEntryFields($carrier)
    {
        return [
            [
                'id' => 'pobocka',
                'label' => $this->l('Affiliate'),
                'placeholder' => $this->l('Search by entering postcode or city name'),
                'required' => true,
                'type' => 'select',
                'select' => [
                    'type' => 'ajax'
                ]
            ]
        ];
    }

    /**
     * @param array $carrier
     *
     * @return boolean
     * @throws PrestaShopException
     */
    function handlesCarrier($carrier)
    {
        return !is_null($this->getCountryEdition($carrier));
    }

    /**
     * @param Cart $cart
     * @param array $carrier
     * @param array $data
     *
     * @return bool
     * @throws PrestaShopException
     */
    function processData(Cart $cart, $carrier, $data)
    {
        $pobocka = (int)$data['pobocka'];
        $cartId = (int)$cart->id;
        $sql = "INSERT INTO " . _DB_PREFIX_ . "shaim_zasilkovna_data(id_cart, id) VALUES ($cartId, $pobocka) ON DUPLICATE KEY UPDATE id = VALUES(id)";
        return Db::getInstance()->execute($sql);
    }


    /**
     * @param string $str
     * @return string
     */
    public function l($str)
    {
        return Translate::getModuleTranslation('chex', $str, 'carrier.zasilkovna');
    }

    /**
     * @param string $field
     * @param string $search
     * @param array $carrier
     *
     * @return array
     * @throws PrestaShopException
     */
    function getSelectOptions($field, $search, $carrier)
    {
        $countryEdition = $this->getCountryEdition($carrier);
        if ($countryEdition) {
            return $this->searchZasilkovna($search, $countryEdition);
        }
        return [];
    }

    /**
     * @param string $search string to search
     * @param string $countryEdition - cz or sk
     * @return array
     * @throws PrestaShopException
     * @throws PrestaShopDatabaseException
     */
    private function searchZasilkovna($search, $countryEdition)
    {
        $sql = (new DbQuery())
            ->select('alias, adresa, psc, id')
            ->from('shaim_zasilkovna')
            ->where('`exists` = 1')
            ->where("`type` = '" .pSQL($countryEdition) . "'")
            ->limit(50, 0);

        if ($search) {
            $search = pSQL($search);
            $sql->where('`psc` LIKE "%' . $search . '%" OR `alias` LIKE "%' . $search . '%" OR `adresa` LIKE "%' . $search . '%" OR `id` = "' . $search . '"');
            $sql->orderBy('(`id` = "'.$search.'") DESC');
        }
        $sql->orderBy('naz_prov ASC');

        $ret = [];
        foreach (Db::getInstance()->executeS($sql) as $row) {
            $ret[] = [
                'value' => $row['id'],
                'label' => $row['alias'] . ': ' . $row['adresa'] . ', ' . $row['psc']
            ];
        }
        return $ret;
    }

    /**
     * @param array $carrier
     * @return string|null
     * @throws PrestaShopException
     */
    private function getCountryEdition($carrier)
    {
        $id = (int)$carrier['id_reference'];
        $dopravceCz = (int)Configuration::get('shaim_zasilkovna_dopravce_cz');
        if ($id === $dopravceCz) {
            return 'cz';
        }
        $dopravceSk = (int)Configuration::get('shaim_zasilkovna_dopravce_sk');
        if ($id === $dopravceSk) {
            return 'sk';
        }
        return null;
    }
}
